"""
The 'textFmt' function provides quick formatting of text. This is
motivated by the fact that Python strings spanning multiple lines contains
the linebreaks verbatim, which is almost never desired.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


def textFmt(*args: Any, **kwargs: Any) -> str:
  """
  Formats a multi-line string by replacing all continuous whitespace,
  including new lines, with a single space. Additionally, intended line
  breaks and indentations may be specified using '<br>' and '<tab>'
  respectively as indicating tokens.

  Args:
    *args: Any number of 'str' objects to be concatenated and formatted.
    Please note that any 'arg' not an instance of 'str' will be replaced
    with 'str(arg)'.
  Kwargs: Additional formatting options:
      - 'newLineToken': Use a different token to indicate line breaks,
      defaults to '<br>'.
      - 'tabToken': Use a different token to indicate indentation,
      defaults to '<tab>'.
      - 'newLineSymbol': Symbol in the returned string indicating new
      line. By default, the value at 'os.linesep' is used. On linux,
      this value is '\n' and on Windows it is '\r\n'.
      - 'indentSymbol': Symbol in the returned string indicating one level
      of indentation, defaults to two spaces.

  Usage:
    from __future__ import annotations

    import sys

    def main(*args) -> int:
      #  Example script
      exampleText = \"\"\"This is the first line,
      but we are still here! <br>But now we are on the second line!
      still here! <br><tab>Third line is even indented!\"\"\"
      formattedText = textFmt(exampleText)
      print(formattedText)

    if __name__ == '__main__':
      sys.exit(main(*sys.argv[1:]))

    output:
   |  This is the first line, but we are still here!
   |  But now we are on the second line!
   |    Third line is even indented!
  """
  if not args:
    return ''
  words = []
  for arg in args:
    if isinstance(arg, str):
      if arg:
        words.append(arg)
      continue
    words.append(str(arg))
  else:
    if not words:
      return ''
  #  Specify tokens and symbols
  nLIn = kwargs.get('newLineToken', '<br>')
  tabIn = kwargs.get('tabToken', '<tab>')
  newTemp = '{{NEWLINE}}'
  tabTemp = '{{TAB}}'
  nLOut = kwargs.get('newLineSymbol', os.linesep)
  tabOut = kwargs.get('indentSymbol', '  ')
  #  Replace tokens with temporary symbols
  parts = [a.replace(nLIn, newTemp).replace(tabIn, tabTemp) for a in words]
  #  Join the parts into a single string
  text = ' '.join(str(part) for part in parts)
  #  Replace multiple spaces with a single space
  text = ' '.join(text.split())
  #  Replace newlines and tabs with the appropriate symbols
  return text.replace(newTemp, nLOut).replace(tabTemp, tabOut).strip()



"""
The 'worktoy.utilities' module provides small, standalone utilities used
across the 'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._class_body_template import ClassBodyTemplate
from ._bipartite_matching import bipartiteMatching
from ._unpack import unpack
from ._maybe import maybe
from ._text_fmt import textFmt
from ._string_list import stringList
from ._directory import Directory
from ._join_words import joinWords
from ._word_wrap import wordWrap

__all__ = [
    'ClassBodyTemplate',
    'bipartiteMatching',
    'unpack',
    'maybe',
    'textFmt',
    'stringList',
    'Directory',
    'joinWords',
    'wordWrap',
]



"""ProtectedError is raised to indicate an attempt to delete a protected
object. For example, a descriptor class could implement the '__delete__'
method to always raise this exception. This provides a more detailed
error. Particularly because both TypeError and AttributeError are being
suggested by large language models. Neither of which is wrong, but lacks
the specificity of this exception.

The ProtectedError class inherits from both TypeError and AttributeError,
ensuring that it is caught in exception clauses pertaining to either.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import DescriptorException
from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class ProtectedError(DescriptorException):
  """ProtectedError is raised to indicate an attempt to delete a protected
  object. For example, a descriptor class could implement the '__delete__'
  method to always raise this exception. This provides a more detailed
  error. Particularly because both TypeError and AttributeError are being
  suggested by large language models. Neither of which is wrong, but lacks
  the specificity of this exception."""

  __slots__ = ('instance', 'desc', 'oldVal')

  def __init__(self, instance: Any, desc: Any, oldValue: Any = None) -> None:
    """Initialize the ReadOnlyError."""
    self.instance = instance
    self.desc = desc
    self.oldVal = oldValue
    DescriptorException.__init__(self, )

  def __str__(self, ) -> str:
    """
    Return the string representation of the ProtectedError.
    """
    oldValue = self.oldVal
    desc = self.desc
    fieldOwner = getattr(self.instance, '__field_owner__', None)
    fieldName = getattr(desc, '__field_name__', None)
    infoSpec = """Attempted to delete protected attribute '%s.%s' 
      with value: '%s'"""
    info = infoSpec % (fieldOwner, fieldName, str(oldValue))
    return textFmt(info)

  __repr__ = __str__



"""
The 'worktoy.utilities.mathematics' module provides mathematical utilities
used across the 'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._constants import pi, arcTan, atan2
from ._trig import sin, cos, tan, cosh, sinh, tanh
from ._exp_log import e, log, exp
from ._factorial import factorial

__all__ = [
    'e',
    'pi',
    'arcTan',
    'atan2',
    'log',
    'exp',
    'sin',
    'cos',
    'tan',
    'cosh',
    'sinh',
    'tanh',
    'factorial',
]



"""
The 'worktoy.ezdata' package provides the EZData dataclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen

from ._data_field import DataField
from ._ez_hook import EZSpaceHook
from ._ez_space import EZSpace
from ._ez_meta import EZMeta
from ._ez_data import EZData

__all__ = [
    'DataField',
    'EZSpaceHook',
    'EZSpace',
    'EZMeta',
    'EZData',
]



"""DataField represents an entry in the EZData classes. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.waitaminute import TypeException


class DataField:
  """DataField represents an entry in the EZData classes. """
  __slots__ = ('key', 'type_', 'val')

  def __init__(self, key: str, type_: type, val: object = None) -> None:
    """Initialize the DataField object."""
    if not isinstance(key, str):
      raise TypeException('key', key, str)
    if not isinstance(type_, type):
      raise TypeException('type_', type_, type)
    if not isinstance(val, type_):
      if val is None:
        try:
          val = type.__call__(type_, )
        except (TypeError, ValueError) as exc:
          raise TypeException('val', val, type_) from exc
      else:
        raise TypeException('val', val, type_)
    self.key = key
    self.type_ = type_
    self.val = val

  def __str__(self) -> str:
    """Get the string representation of the DataField object."""
    infoSpec = """%s: %s = %s(%s)"""
    clsName = type(self).__name__
    typeName = self.type_.__name__
    return infoSpec % (self.key, typeName, clsName, self.val)

  def __repr__(self) -> str:
    """
    Get the code that would create this DataField object.
    """
    infoSpec = """%s(%r, %r, %r)"""
    clsName = type(self).__name__
    typeName = self.type_.__name__
    return infoSpec % (clsName, self.key, typeName, self.val)

  def __eq__(self, other: object) -> bool:
    """
    Compares other against self.val or if other is also a 'DataField'
    object, compares against self.key, self.type_, and self.val.
    """
    if type(self) is type(other):
      if TYPE_CHECKING:  # pragma: no cover
        assert isinstance(other, DataField)
      if self is other:
        return True
      if self.key != other.key:
        return False
      if self.type_ != other.type_:
        return False
      if self.val != other.val:
        return False
      return True
    if isinstance(other, self.type_):
      return True if self.val == other else False
    return NotImplemented

  def __hash__(self) -> int:
    """
    Returns the hash of the DataField object.
    """
    return hash((self.key, self.type_.__name__, self.val))



"""
The 'worktoy.waitaminute' package provides the custom exceptions used
across the 'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import desc
from . import meta
from . import dispatch
from ._attribute_error_factory import attributeErrorFactory
from ._type_exception import TypeException
from ._write_once_error import WriteOnceError
from ._missing_variable import MissingVariable
from ._variable_not_none import VariableNotNone
from ._unpack_exception import UnpackException
from ._path_syntax_exception import PathSyntaxException
from ._subclass_exception import SubclassException

___all__ = [
    'desc',
    'meta',
    'dispatch',
    'attributeErrorFactory',
    'TypeException',
    'WriteOnceError',
    'MissingVariable',
    'VariableNotNone',
    'UnpackException',
    'PathSyntaxException',
    'SubclassException',
]



"""
KeeSpace provides the namespace object for the KeeMeta metaclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls import AbstractNamespace
from . import KeeSpaceHook  # Private to module

if TYPE_CHECKING:  # pragma: no cover
  pass


class KeeSpace(AbstractNamespace):
  """
  KeeSpace provides the namespace object for the KeeMeta metaclass.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public variables
  keeHook = KeeSpaceHook()



"""IllegalInstantiation is a custom exception raised to indicate than
an attempt was made to instantiate a class under illegal conditions."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class IllegalInstantiation(TypeError):
  """IllegalInstantiation is a custom exception raised to indicate than
  an attempt was made to instantiate a class under illegal conditions."""

  __slots__ = ('cls',)

  def __init__(self, cls_: type) -> None:
    """Initialize the IllegalInstantiation with the class."""
    self.cls = cls_
    TypeError.__init__(self, )

  def __str__(self, ) -> str:
    """Return the string representation of the IllegalInstantiation
    object."""
    clsName = self.cls.__name__
    infoSpec = """Illegal instantiation of class '%s'"""
    info = infoSpec % clsName
    return textFmt(info)

  __repr__ = __str__



"""
DescriptorException provides a base class for the exceptions raised by
descriptors. By having a shared base class, the 'Object' class is able to
recognize situations where a descriptor has raised an exception. These
exceptions propagate *without* being passed to the fallback method.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


class DescriptorException(Exception):
  """
  Base class for exceptions raised by descriptors.
  """
  pass



"""The 'worktoy.static' module provides low level parsing and casting
utilities. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from ._type_cast import typeCast
from ._pre_class import PreClass
from ._type_sig import TypeSig
from ._dispatch import Dispatch
from ._overload import overload

__all__ = [
    'typeCast',
    'PreClass',
    'TypeSig',
    'Dispatch',
    'overload',
]



"""
MetaType provides the meta-metaclass for the 'worktoy' library. All
metaclasses used across the library both derive from and base on this
class. This is necessary to prevent metaclass conflicts.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Type, TypeAlias, Union, Self

  Bases: TypeAlias = tuple[type, ...]
  Namespace: TypeAlias = dict[str, Any]
  NamespaceClass: TypeAlias = Union[Self, Type[Namespace]]

  Meta: TypeAlias = Type[type]
  MetaMeta: TypeAlias = Type[Meta]


class _Space:
  """
  Private descriptor class providing the namespace for the namespace
  object class used by the metaclass.
  """

  __field_name__ = None
  __field_owner__ = None

  def __get__(self, mcls: type, mmcls: type, ) -> Any:
    if mcls is None:
      if mmcls is self.__field_owner__:
        return self
      return self.__get__(mmcls, mmcls)
    testSpace = mcls.__prepare__('test', (), )
    return type(testSpace)

  def __set_name__(self, owner: type, name: str) -> None:
    """
    Sets the name of the descriptor in the owner class.
    """
    self.__field_name__ = name
    self.__field_owner__ = owner


class MetaType(type):
  """
  MetaType provides the meta-metaclass for the 'worktoy' library. All
  metaclasses used across the library both derive from and base on this
  class. This is necessary to prevent metaclass conflicts.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __namespace_class__ = None

  #  Virtual Variables
  namespaceClass = _Space()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __str__(cls, ) -> str:
    """Returns the name of the class. """
    return """%s[metaclass=%s]""" % (cls.__name__, cls.__class__.__name__)

  __repr__ = __str__



"""
EZData leverages the 'worktoy' library to provide a dataclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import EZMeta

if TYPE_CHECKING:  # pragma: no cover
  pass

Func = type('_', (type,), dict(__instancecheck__=callable))('_', (), {})


def trustMeBro(callMeMaybe: Func) -> Func:
  """
  This is a decorator that can be used to mark a function as a root
  function in the EZData class.
  """
  callMeMaybe.__is_root__ = True
  return callMeMaybe


class EZData(metaclass=EZMeta):
  """
  EZData is a dataclass that provides a simple way to define data
  structures with validation and serialization capabilities.
  """
  pass

  @trustMeBro
  def __init__(self, *args, **kwargs) -> None:
    """
    Here for type hinting purposes only!
    """

  @trustMeBro
  def __len__(self, *args, **kwargs) -> None:
    """
    Here for type hinting purposes only!
    """



"""
ReservedNames provides a list of reserved names that are set
automatically by the interpreter.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt
from ...core import Object

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Iterator


class ReservedNames(Object):
  """
  ReservedNames provides a list of reserved names that are set
  automatically by the interpreter.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __reserved_names__ = [
      '__dict__',
      '__weakref__',
      '__module__',
      '__annotations__',
      '__match_args__',
      '__doc__',
      '__name__',
      '__qualname__',
      '__firstlineno__',
      '__static_attributes__',
  ]

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __iter__(self) -> Iterator[str]:
    """
    Iterate over the reserved names.
    """
    yield from self.__reserved_names__

  def __contains__(self, name: str) -> bool:
    """Check if the name is in the reserved names."""
    for item in self:
      if item == name:
        return True
    return False

  def __len__(self) -> int:
    """Get the number of reserved names."""
    return len(self.__reserved_names__)

  def __str__(self, ) -> str:
    """Get the string representation of the metaclass."""
    info = 'ReservedNames:\n%s'
    names = '<br><tab>'.join([name for name in self])
    return textFmt(info % names)

  __repr__ = __str__

  def __instance_get__(self, *args, **kwargs) -> Any:
    """
    Get the reserved name by its string identifier.
    """
    return self



"""EZSpace provides the namespace for the EZData class. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import DataField, EZSpaceHook
from ..mcls import BaseSpace
from ..utilities import maybe


class EZSpace(BaseSpace):
  """
  EZSpace provides the namespace for the EZData class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __data_fields__ = None

  #  Public Variables
  ezHook = EZSpaceHook()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getDataFields(self) -> list[DataField]:
    """Get the data fields."""
    out = maybe(self.__data_fields__, [])
    bases = [*self.getBases(), ]
    for base in bases:
      space = getattr(base, '__namespace__', None)
      fields = getattr(space, '__data_fields__', None)
      if fields is None:
        continue
      for field in fields:
        if field.key not in [f.key for f in out]:
          out.append(field)
    return out

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addField(self, key: str, type_: type, val: object) -> None:
    """Add a field to the data fields."""
    if self.__data_fields__ is None:
      self.__data_fields__ = []
    dataField = DataField(key, type_, val)
    existing = self.getDataFields()
    self.__data_fields__ = [*existing, dataField]



"""
The 'unpack' function squeezes a tuple of arguments until no array like
member remains.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


def unpack(*args: Any, **kwargs) -> tuple[Any, ...]:
  """
  Recursively unpacks iterables from positional arguments.

  This function traverses the given positional arguments and flattens
  any array-like values (iterables), except for str and bytes, which
  are treated as atomic.

  Supported keyword arguments:

  - shallow (bool, default False):
    If True, only unpacks the first level of nested iterables.

  - strict (bool, default True):
    If True, raises ValueError if no iterable argument was found.
    If False, returns arguments unchanged if no iterable is found.

  Args:
    *args: Positional arguments to unpack.
    **kwargs: Optional 'shallow' and 'strict' flags.

  Returns:
    tuple[Any, ...]: A flattened tuple of arguments.

  Raises:
    ValueError: If strict is True and no iterable was found.
  """
  if not args:
    if kwargs.get('strict', True):
      from worktoy.waitaminute import UnpackException
      raise UnpackException(*args, )
    return ()
  out = []
  iterableFound = False
  for arg in args:
    if isinstance(arg, (str, bytes,)):
      out.append(arg)
      continue
    if isinstance(arg, Iterable):
      if kwargs.get('shallow', False):
        out.extend(arg)
      else:
        out = [*out, *unpack(*arg, shallow=False, strict=False)]
      iterableFound = True
      continue
    out.append(arg)
  if iterableFound or not kwargs.get('strict', True):
    return (*out,)
  from worktoy.waitaminute import UnpackException
  raise UnpackException(*args, )



"""
Sentinel provides the base class for the sentinel objects provided by the
'worktoy.core.sentinels' module. By subclassing 'type' and becoming a
metaclass, it allows sentinel objects to be instances of 'type' and thus
classes.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.utilities import maybe

if TYPE_CHECKING:  # pragma: no cover
  from typing import Self, Never

  Bases = tuple[type, ...]


class _Sentinel(type):
  """
  Sentinel keeps track of all sentinel objects and prevents multiple
  sentinels of the same name from being created. It also prevents its
  derived classes from being instantiated.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __registered_sentinels__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _getRegisteredSentinels(mcls) -> list[Self]:
    """
    Get the dictionary of registered sentinel classes.
    """
    return maybe(mcls.__registered_sentinels__, [])

  @classmethod
  def _getNamedSentinel(cls, sentinelName: str, ) -> Self:
    """
    Returns registered sentinel having this name.
    """
    existing = cls._getRegisteredSentinels()
    for existingSentinel in existing:
      if existingSentinel.__name__ == sentinelName:
        return existingSentinel
    raise KeyError(sentinelName)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _registerSentinel(mcls, sentinel: Self) -> None:
    """
    Register a sentinel class to prevent multiple sentinels of the same name
    from being created.
    """
    existing = mcls._getRegisteredSentinels()
    mcls.__registered_sentinels__ = [*existing, sentinel]

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> dict:
    """
    Prepare the class namespace for the sentinel class.
    """
    return dict()

  def __new__(mcls, name: str, bases: Bases, space: dict, **kwargs) -> Self:
    """
    Create a new sentinel class. If a sentinel with the same name already
    exists, return it instead of creating a new one.
    """
    try:
      cls = mcls._getNamedSentinel(name)
    except KeyError as keyError:
      if kwargs.get('_recursion', False):
        raise keyError from RecursionError
      cls = super().__new__(mcls, name, (), {}, **kwargs)
      mcls._registerSentinel(cls)
      return mcls.__new__(mcls, name, (), {}, _recursion=True)
    else:
      return cls

  def __call__(cls, *__, **_) -> Never:
    """
    Raises 'IllegalInstantiation'.
    """
    from worktoy.waitaminute.meta import IllegalInstantiation
    raise IllegalInstantiation(cls)


class Sentinel(metaclass=_Sentinel):
  """
  Sentinel is the base class for all sentinel objects in the 'worktoy.core'
  module. It prevents instantiation and ensures that only one instance of
  each sentinel exists.
  """

  __slots__ = ()



"""
Object provides the most basic object used by the 'worktoy' library. It
stands in for the 'object' type by adding functionality that must be
shared by every object in the library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import re
from typing import TYPE_CHECKING

from ..utilities import Directory, maybe
from . import ContextInstance, MetaType
from .sentinels import THIS, DESC, OWNER, DELETED, LOCKED
from ..waitaminute import TypeException, MissingVariable, \
  attributeErrorFactory
from ..waitaminute.desc import AccessError

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Self, Optional, Type, TypeAlias
  from types import TracebackType

  ExcType: TypeAlias = Optional[Type[Exception]]
  ExcVal: TypeAlias = Optional[Exception]
  Trace: TypeAlias = Optional[TracebackType]


class Object(object, metaclass=MetaType):
  """
  Object provides the most basic object used by the 'worktoy' library. It
  stands in for the 'object' type by adding functionality that must be
  shared by every object in the library.

  The class implements '__init__' that collects and holds references to
  the positional and keyword arguments passed to the constructor. These
  are later available through the 'getPosArgs' and 'getKeyArgs' methods.
  The class also implements '__set_name__' used by the descriptor protocol
  to inform instances defined in class bodies when the owning class is
  created and the name by which they appear in the class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables

  #  Private Variables
  __field_owner__ = None
  __field_name__ = None
  __pos_args__ = None
  __key_args__ = None
  __context_instance__ = None
  __context_owner__ = None
  __call_chain__ = []

  #  Public Variables
  directory = Directory()
  instance = ContextInstance()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getFieldOwner(self) -> Optional[type]:
    """Getter for the field owner of the descriptor object"""
    if self.__field_owner__ is None:
      from ..waitaminute import MissingVariable
      raise MissingVariable('__field_owner__', type)
    return self.__field_owner__

  def getFieldName(self) -> Optional[str]:
    """Getter for the field name of the descriptor object"""
    if self.__field_name__ is None:
      from ..waitaminute import MissingVariable
      raise MissingVariable('__field_name__', str)
    return self.__field_name__

  def getPosArgs(self, **kwargs) -> tuple[Any, ...]:
    """Getter for the positional arguments of the object."""
    if self.__pos_args__ is None:
      return ()
    if not isinstance(self.__pos_args__, (list, tuple)):
      from ..waitaminute import TypeException
      raise TypeException('__pos_args__', self.__pos_args__, list, tuple, )
    thisObj = maybe(kwargs.get('THIS', self.__context_instance__), THIS)
    ownerObj = maybe(kwargs.get('OWNER', self.__context_owner__), OWNER)
    descObj = maybe(kwargs.get('DESC', self), )
    out = []
    for arg in self.__pos_args__:
      if arg is THIS:
        out.append(thisObj)
        continue
      if arg is OWNER:
        out.append(ownerObj)
        continue
      if arg is DESC:
        out.append(descObj)
        continue
      out.append(arg)
    return (*out,)

  def getKeyArgs(self, **kwargs) -> dict[str, Any]:
    """Getter for the keyword arguments of the object."""
    if self.__key_args__ is None:
      return dict()
    if not isinstance(self.__key_args__, dict):
      from ..waitaminute import TypeException
      raise TypeException('__key_args__', self.__key_args__, dict, )
    thisObj = maybe(kwargs.get('THIS', self.__context_instance__), THIS)
    ownerObj = maybe(kwargs.get('OWNER', self.__context_owner__), OWNER)
    descObj = maybe(kwargs.get('DESC', self), )
    out = dict()
    for key, value in self.__key_args__.items():
      if value is THIS:
        out[key] = thisObj
        continue
      if value is OWNER:
        out[key] = ownerObj
        continue
      if value is DESC:
        out[key] = descObj
        continue
      out[key] = value
    return {**out, }

  def getContextInstance(self) -> Any:
    """Returns the contextual instance or raises 'WithoutException'"""
    if self.hasContext():
      return self.__context_instance__
    from ..waitaminute.desc import WithoutException
    raise WithoutException(self)

  def hasContext(self) -> bool:
    """
    Returns True if the descriptor has a context, i.e. if it has been
    created with 'createContext' and not exited with 'exitContext'.
    """
    own = self.__context_owner__
    ins = self.__context_instance__
    if own is LOCKED or own is None:
      return False
    return True

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args: Any, **kwargs: Any) -> None:
    object.__init__(self)
    self.__pos_args__ = args
    self.__key_args__ = kwargs
    self.__context_instance__ = LOCKED
    self.__context_owner__ = LOCKED

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: Type[object], name: str) -> None:
    self.__field_owner__ = owner
    self.__field_name__ = name

  def __get__(self, instance: Any, owner: type) -> Any:
    """
    Returns the root of the descriptor owning the hook.
    """
    if instance is None:
      return self
    with self.createContext(instance, owner) as context:
      value = context.__instance_get__()
    return self._deletedGuard(instance, value)

  def __set__(self, instance: Any, newValue: Any, **kwargs) -> None:
    """
    Sets the value of the descriptor in the instance. If accessing an
    attribute would raise an exception, it should not prevent setting a
    value on that attribute. Since the 'setter' control flow
    """
    with self.createContext(instance, type(instance)) as context:
      context.__instance_set__(newValue, **kwargs)

  def __delete__(self, instance: Any, **kwargs) -> None:
    """
    Deletes the value of the descriptor in the instance.
    """
    try:
      with self.createContext(instance, type(instance)) as context:
        oldVal = self._deletedGuard(instance, context.__instance_get__())
        context.__instance_delete__(oldVal, **kwargs)
    except AccessError as accessError:
      ownerName = type(instance).__name__
      fieldName = getattr(self, '__field_name__', 'Unknown')
      attributeError = attributeErrorFactory(ownerName, fieldName)
      raise attributeError from accessError

  def __enter__(self, ) -> Self:
    """
    Must be used with along with 'enterContext(instance, owner)'.
    """
    if self.hasContext():
      return self
    from ..waitaminute.desc import WithoutException
    raise WithoutException(self)

  def __exit__(self, _, exception: BaseException, __) -> None:
    """
    Must be used with along with 'exitContext'.
    """
    try:
      if exception is not None:
        raise exception
    except MissingVariable as missingVariable:
      raise AttributeError from missingVariable
    finally:
      self.exitContext()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, *args, **kwargs) -> Any:
    """
    Subclasses should implement this method to define the
    instance-specific getter. The root instance is accessed through the
    'root' attribute.
    """
    return self

  def __instance_set__(self, value: Any, *args, **kwargs) -> None:
    """
    Subclasses should implement this method to define the
    instance-specific setter. The root instance is accessed through the
    'root' attribute.
    """
    from ..waitaminute.desc import ReadOnlyError
    raise ReadOnlyError(self.instance, self, value)

  def __instance_delete__(self, oldVal: Any, *args, **kwargs) -> None:
    """
    Subclasses should implement this method to define the
    instance-specific deleter. The root instance is accessed through the
    'root' attribute.
    """
    from ..waitaminute.desc import ProtectedError
    raise ProtectedError(self.instance, self, oldVal)

  def createContext(self, instance: Any, owner: type, ) -> Self:
    """
    Creates a context for the descriptor. The context is used to
    store the instance and owner of the descriptor.
    """
    self.__context_instance__ = instance
    self.__context_owner__ = owner
    return self

  def exitContext(self) -> Self:
    """
    Exits the context of the descriptor. The method restores the
    descriptor to its previous state.
    """
    self.__context_instance__ = LOCKED
    self.__context_owner__ = LOCKED
    return self

  def _deletedGuard(self, instance: Any, value: Any) -> Any:
    """
    A guard that raises an exception if the value is 'DELETED'. This is
    used to prevent accessing deleted attributes.
    """
    if value is DELETED:
      infoSpec = """%s: '%s' object has no attribute '%s'."""
      eType = 'AttributeError'
      ownerName = type(instance).__name__
      fieldName = getattr(self, '__field_name__', 'object')
      info = infoSpec % (eType, ownerName, fieldName)
      raise AttributeError(info)
    return value

  def getPrivateName(self, ) -> str:
    """
    Returns the name chain of the descriptor. This is used to access
    the name of the descriptor in the context of the instance.
    """
    fieldName = self.__field_name__
    pattern = re.compile(r'(?<!^)(?=[A-Z])')
    return '__%s__' % pattern.sub('_', fieldName).lower()

  @classmethod
  def parseKwargs(cls, *args, **kwargs) -> tuple[Any, dict]:
    """
    Parses the keyword arguments for value matching keys and types given
    in positional arguments. The return value is a tuple of the value
    found and the remaining keyword arguments. If no value is found,
    the returned tuple will be 'None' and all the keyword arguments
    received. If no types are given, the type of the value is ignored.
    """
    typeArgs, keys = [], []
    for arg in args:
      if isinstance(arg, type):
        typeArgs.append(arg)
        continue
      if isinstance(arg, str):
        keys.append(arg)
        continue
    if not keys:
      return None, {**kwargs, }
    typeArgs = typeArgs or [object, ]
    if complex in typeArgs:
      if float not in typeArgs:
        typeArgs.append(float)
      if int not in typeArgs:
        typeArgs.append(int)
    elif float in typeArgs:
      if int not in typeArgs:
        typeArgs.append(int)
    for key in keys:
      if key in kwargs:
        value = kwargs[key]
        for type_ in typeArgs:
          if isinstance(value, type_):
            del kwargs[key]
            return value, {**kwargs, }
        else:
          raise TypeException(key, value, *typeArgs, )
    else:
      return None, {**kwargs, }



"""HookException is raised from the AbstractNamespace class to wrap
exceptions raised by __getitem__ hooks. This is necessary to avoid
confusion with the expected KeyError exception in the metacall system."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from worktoy.mcls import AbstractNamespace
  from worktoy.mcls.space_hooks import AbstractSpaceHook


class HookException(Exception):
  """
  This custom exception allows get item hooks to interrupt calls to
  __getitem__. Because the metacall system requires the __getitem__ to
  specifically raise a KeyError in certain situations, an exception raised
  by a hook might be confused for the KeyError. Instead,
  the AbstractNamespace class will catch exceptions raised by hooks and
  raise them from this exception:
  For example:
  try:
    hook(self, key, val)
  except Exception as exception:
    raise HookException(exception) from exception
  """

  __slots__ = (
      'initialException',
      'namespaceObject',
      'itemKey',
      'errorValue',
      'hookFunction'
  )

  def __init__(
      self,
      exception: Exception,
      namespace: AbstractNamespace,
      key: str,
      val: object,
      hook: AbstractSpaceHook,
  ) -> None:
    self.initialException = exception
    self.namespaceObject = namespace
    self.itemKey = key
    self.errorValue = val
    self.hookFunction = hook
    Exception.__init__(self, )

  def __str__(self) -> str:
    """
    String representation of the HookException.
    """
    spec = """HookException raised from %s! Key: '%s', Value: '%s', 
    Hook: '%s'! Initial exception: %s"""
    cls = type(self).__name__
    info = spec % (
        self.namespaceObject, self.itemKey, self.errorValue,
        self.hookFunction, self.initialException
    )
    return info

  __repr__ = __str__



"""
This file is part of WorkToy.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.utilities.mathematics import pi

if TYPE_CHECKING:  # pragma: no cover
  from typing import Union


def _sin(x: complex, h: bool = None) -> Union[float, complex]:
  """Returns the sine of x."""
  x = x + 0j
  term = x
  result = x
  n = 1
  while abs(term) > 1e-32:
    term *= (1 if h else -1) * x * x / ((2 * n) * (2 * n + 1))
    result += term
    n += 1
  else:
    return result


def _cos(x: complex, h: bool = None) -> Union[float, complex]:
  """Returns the cosine of x."""
  x = x + 0j
  term = 1
  result = 1
  n = 1
  while abs(term) > 1e-32:
    term *= (1 if h else -1) * x * x / ((2 * n - 1) * (2 * n))
    result += term
    n += 1
  else:
    return result


def cos(x: complex) -> Union[float, complex]:
  """Returns the cosine of x."""
  if abs(x) < 1e-32:
    return 1.
  # if isinstance(x, (float, int)):
  #   if x < 0:
  #     return cos(-x)
  #   if x > pi:
  #     return -cos(x - pi)
  #   if x > pi / 2:
  #     return -cos(pi - x)
  return _cos(x, )


def sin(x: complex) -> Union[float, complex]:
  """Returns the sine of x."""
  if abs(x) < 1e-32:
    return .0
  # if isinstance(x, (float, int)):
  #   if x > 2 * pi:
  #     return sin(x - 2 * pi)
  #   if x < 2 * pi:
  #     return sin(x + 2 * pi)
  #   if x < 0:
  #     return -sin(-x)
  #   if x > pi / 2:
  #     return sin(pi - x)
  return _sin(x, )


def sinh(x: complex) -> Union[float, complex]:
  """Returns the hyperbolic sine of x."""
  return _sin(x, h=True)


def cosh(x: complex) -> Union[float, complex]:
  """Returns the hyperbolic cosine of x."""
  return _cos(x, h=True)


def tan(x: complex) -> Union[float, complex]:
  """Returns the tangent of x."""
  if abs(x) < 1e-16:
    return 0
  if abs(pi - 2 * x) < 1e-16:
    raise ZeroDivisionError
  sinX = _sin(x)
  cosX = _cos(x)
  return sinX / cosX


def tanh(x: complex) -> Union[float, complex]:
  """Returns the hyperbolic tangent of x."""
  if abs(x) < 1e-32:
    return 0
  return sinh(x) / cosh(x)



"""
AbstractMetaclass provides the baseclass for custom metaclasses.
"""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Base
from . import AbstractNamespace as ASpace
from ..core import MetaType
from ..utilities import maybe

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any
else:
  pass


class AbstractMetaclass(MetaType, metaclass=MetaType):
  """
  Abstract base for custom metaclasses that separates concerns between
  class construction and class behavior.

  This design delegates the initial class namespace to a custom object
  returned by '__prepare__', while keeping class semantics within the
  metaclass itself.

  The namespace object may define a method called 'compile()', which
  should return the finalized dictionary to be passed to 'type.__new__'.

  This separation enables:
  - Customizing how the class body is assembled without affecting the
    resulting class behavior.
  - Modifying what it means to be a class, independently of how the
    class body is constructed.

  - CUSTOM NAMESPACE COMPILATION -

  This module also provides an 'AbstractNamespace' class intended to be
  used with '__prepare__'. It defines a 'compile()' method and an
  '__init__()' signature compatible with the arguments passed to
  '__prepare__'.

  from . import AbstractNamespace as ASpace  # For brevity

  Specifically, '__prepare__' is defined as:
    def __prepare__(mcls, name: str, bases: Base, **kwargs) -> ASpace

  And 'ASpace' ('AbstractNamespace') implements:
    def __init__(self, mcls: type, name: str, bases: Base, **kwargs)

  This allows the namespace object to be instantiated with full context
  about the class being defined, while remaining isolated from the
  behavior of the resulting class itself. For more information about the
  class body execution flow, see the 'AbstractNamespace' class
  documentation.

  - CUSTOM CLASS CREATION -

  After assembly and compilation of the namespace, the metaclass itself
  takes responsibility for validating and finalizing the class.

  This includes a pass over the namespace via the '_validateNamespace'
  static method. It performs checks for common mistakes such as typos
  in special method names—for example:
    - '__set_item__' instead of '__setitem__'
    - '__get_attr__' instead of '__getattr__'
    - '__setname__' instead of '__set_name__'

  If any such names are found, a 'QuestionableSyntax' error is raised
  to prompt correction.

    -  '__del__' or '__delete__' ? -
  The descriptor protocol allows classes to define what happens when an
  attribute is deleted from an instance. This is handled by the '__delete__'
  method. It is much less common than '__get__' and '__set__', which govern
  attribute access and assignment, respectively.

  Because of the naming similarity to '__del__'—a special method for object
  finalization—it's easy to accidentally implement '__del__' when one meant
  '__delete__'.

  Bugs caused by incorrect use of `__del__`—especially when accidentally
  used instead of `__delete__`—are notoriously difficult to trace. Since
  `__del__` is called only when the object is garbage collected (which may
  be delayed or never happen), the consequences of the mistake are often
  deferred until long after the original action that should have triggered
  cleanup. This makes it extremely hard to correlate the broken behavior
  with its source. Worse still, because `__del__` doesn’t raise errors if
  used in the wrong context, failures are often silent, leading to
  inconsistent state, memory leaks, or subtle bugs in object lifecycles that
  resist even thorough debugging.

  For the above reasons, the 'worktoy' library will raise 'SyntaxError'
  whenever '__del__' is found in the namespace. If an implementation of
  '__del__' is actually intended, the class creation must be invoked
  with the keyword argument 'trustMeBro=True'.

    - Standard Methods -
  While not implemented in this metaclass, the following pattern allows
  sub-metaclasses to implement automatically generated methods in certain
  cases. The 'worktoy.ezdata.EZMeta' dataclass implementation exemplifies
  this pattern. Derived classes can define properties directly in the
  class body, and 'EZMeta' will automatically generate the necessary
  methods, such as '__init__', allowing for instantiation with either
  positional, keyword or even mixed arguments.

  The sub-metaclass would implement factories as static methods:

  from types import FunctionType as Func

  Bases: TypeAlias = tuple[type, ...]
  Space: TypeAlias = dict[str, Any]

  @staticmethod
  def __init_factory__(name: str, bases: Bases, space: Space, **kw) -> Func

  @staticmethod
  def __new_factory__(name: str, bases: Bases, space: Space, **kw) -> Func

  @staticmethod
  def __str_factory__(name: str, bases: Bases, space: Space, **kw) -> Func

  Finally, the sub-metaclass would have to implement:

  @staticmethod
  def autoGenMethods(name: str, bases: Bases, space: Space, **kw) -> Space:
    This method would be invoked after the namespace has been validated
    and before the class is created. It would return a modified namespace
    with the necessary methods added.

    - Notifying Baseclasses -
  Having validated the namespace, the metaclass falls back to type.__new__
  and returns the newly created class. Finally, this class arrives in the
  '__init__' method where the metaclass notifies any baseclass that
  implements the '__subclasshook__' method of the class creation. This
  marks the end of the class creation process.

  - CUSTOM CLASS BEHAVIOR -

  Once a class has been created by the metaclass system, it may define
  its own runtime behavior by implementing special methods prefixed with
  `__class_`. These allow the class object itself to participate directly
  in common operations such as being called, iterated, or printed.

  These methods resemble `__class_getitem__` from standard Python but are
  more general. Each of them overrides a specific class-level behavior.

  The following hooks relate to common class-level operations:

  - __class_call__(cls, *args, **kwargs) -> Any
    Called when the class object is called like a function. Overrides the
    default behavior of constructing instances. Can be used to implement
    singletons, factories, registries, etc.

  - __class_instancecheck__(cls, obj: Any) -> bool
    Called during isinstance(obj, cls). Controls how instance membership
    is determined. Supersedes metaclass-level __instancecheck__.

  - __class_subclasscheck__(cls, sub: type) -> bool
    Called during issubclass(sub, cls). Controls dynamic subclass logic.
    Allows behavior similar to abstract base classes or trait systems.

  #  The following hooks allow classes to define how they are printed

  - __class_str__(cls) -> str
    Called when str(cls) is invoked. Provides human-readable string form
    for dynamically generated or aliased classes.

  - __class_repr__(cls) -> str
    Called when repr(cls) is invoked. Allows classes to override their
    debug representation.

  #  The following hooks relate to class-level iteration

  - __class_iter__(cls) -> Iterator
    Called when iter(cls) is invoked. Makes the class object iterable.
    Useful for registry-style classes, enums, and similar patterns.

  - __class_next__(cls) -> Any
    Called when next(cls) is invoked. Meaningful only if the class itself
    is its own iterator as returned by __class_iter__.

  - __class_bool__(cls) -> bool
    Called when bool(cls) is invoked. Allows classes to define their truth
    value. By default, every class is 'truthy'.

  - __class_contains__(cls, item: Any) -> bool
    Allows classes to define membership checks on the class level. By
    default, this checks if the item is an instance of the class itself.

  - __class_len__(cls) -> int
    Called when len(cls) is invoked.

  - __class_hash__(cls) -> int
    Called when hash(cls) is invoked. Allows classes to define their own
    hash value. Defaults to:
    mcls = type(cls)  # The metaclass of the class
    baseNames = [b.__name__ for b in cls.__bases__]
    return hash((cls.__name__, *baseNames, mcls.__name__))
    #  PLEASE NOTE: The 'overload' protocol provided by the 'worktoy'
    library expects this exact hash value. Reimplementing the hash value
    will make the dispatching of overloads unable to 'fast' recognize the
    class.

  - __class_eq__(cls, other: Any) -> bool
    Called to allow classes to equal each other. Please note that this
    inclusion is for completeness more than anything else. The '__eq__' in
    this metaclass does look for '__class_eq__' on the class, but falls
    back to __class_hash__.

  The following hooks allows dictionary-like access to the class.

  - __class_getitem__(cls, item: Any) -> Any
    Called when cls[item] is invoked. Please note that this is already
    implemented in Python 3.7+ as a standard class method. It is listed
    here only for completeness. This means that this metaclass does not
    need to implement '__getitem__' to look for the '__class_getitem__' on
    the class itself. In fact, the __getitem__ on the metaclass would only
    ever be invoked if Foo['bar'] is invoked on a class Foo that does not
    implement '__class_getitem__'.

  - __class_setitem__(cls, item: Any, value: Any) -> None
    Called when cls[item] = value is invoked. 
    
  - __class_delitem__(cls, item: Any) -> None
    Called when del cls[item] is invoked.

    - Class Attribute Hooks -

  - __class_getattr__(name: str, exception: Exception) -> Any
    If a non-existing attribute is attempted accessed on a class object,
    the '__getattr__' method on the metaclass is invoked. This method
    allows the class itself to handle this case. It is strongly advised
    that this method, and '__getattr__' in general, raises an
    AttributeError unless the key passed to it has a valid and sensible
    meaning in the context.

  The following hooks allow classes to define custom behavior for
  attribute assignment and deletion at the class level.

  - __class_setattr__(name: str, value: Any) -> None
  - __class_delattr__(name: str) -> None

  The following hooks would be relevant only for nested classes that
  implement the descriptor protocol. These class-level descriptor hooks
  remain unimplemented due to unresolved hazards in Python's class
  construction behavior. In particular, referencing other class objects
  while a metaclass is "awake" (i.e., inside its __prepare__, __new__,
  or __init__) can lead to context leakage. Python may interpret unrelated
  class references within the scope of the active metaclass, sometimes
  routing calls to the wrong metaclass entirely.

  - __class_get__(cls, instance: Any, owner: type) -> Any
  - __class_set__(cls, instance: Any, value: Any) -> None
  - __class_delete__(cls, instance: Any) -> None
  - __class_set_name__(cls, owner: type, name: str) -> None

  Finally, the following hooks are logically meaningless:

  - __class_init__(...)
    There is no sensible point at which this would be called. Class
    objects are constructed by metaclasses, and no outer context exists
    in which to simulate an “init” phase.

  - __class_new__(...)
    Similar to above, class creation is driven by the metaclass’s
    __new__, and there is no standard mechanism for a class to override
    or participate in that step on its own behalf.

  - __class_del__(...)
    This remains unimplemented for the same reason as why the namespace
    validator described above raises a SyntaxError when it encounters
    '__del__' in the namespace.

  - __class_getattribute__(...)
    [REDACTED: Cognito Hazard]
  """

  @classmethod
  def __prepare__(mcls, name: str, bases: Base, **kwargs) -> ASpace:
    """The __prepare__ method is invoked before the class is created. This
    implementation ensures that the created class has access to the safe
    __init__ and __init_subclass__ through the BaseObject class in its
    method resolution order."""
    return ASpace(mcls, name, bases, **kwargs)

  def __new__(mcls, name: str, bases: Base, space: ASpace, **kw) -> type:
    """The __new__ method is invoked to create the class."""
    if hasattr(space, 'compile'):
      namespace = space.compile()
    else:
      namespace = mcls.__prepare__(name, bases, **kw)
      namespace = namespace.compile(space)
    cls = MetaType.__new__(mcls, name, bases, namespace, **kw)
    if hasattr(space, 'getHooks'):
      for hook in space.getHooks():
        cls = maybe(hook.newClassPhase(cls), cls)
    return cls

  def __init__(cls, name: str, bases: Base, space: ASpace, **kwargs) -> None:
    """The __init__ method is invoked to initialize the class."""
    try:
      cls.__class_init__(name, bases, space, **kwargs)
    except NotImplementedError:
      pass
    cls._notifySubclassHook(cls, *bases)

  def __call__(cls, *args, **kwargs) -> Any:
    try:
      out = cls.__class_call__(*args, **kwargs)
    except NotImplementedError:
      return MetaType.__call__(cls, *args, **kwargs)  # NOQA
    else:
      return out

  def __instancecheck__(cls, obj: Any) -> bool:
    try:
      out = cls.__class_instancecheck__(obj)
    except NotImplementedError:
      return MetaType.__instancecheck__(cls, obj)  # NOQA
    else:
      return True if out else False

  def __subclasscheck__(cls, subclass: type) -> bool:
    try:
      out = cls.__class_subclasscheck__(subclass)
    except NotImplementedError:
      return MetaType.__subclasscheck__(cls, subclass)  # NOQA
    else:
      return True if out else False

  def __str__(cls) -> str:
    try:
      out = cls.__class_str__()
    except NotImplementedError:
      return MetaType.__str__(cls)  # NOQA
    else:
      return out

  def __repr__(cls) -> str:
    try:
      out = cls.__class_repr__()
    except NotImplementedError:
      return MetaType.__repr__(cls)  # NOQA
    else:
      return out

  def __iter__(cls) -> Any:
    try:
      out = cls.__class_iter__()
    except NotImplementedError:
      infoSpec = """type object '%s' is not iterable"""
      info = infoSpec % type(cls).__name__
      raise TypeError(info)
    else:
      return out

  def __next__(cls) -> Any:
    try:
      out = cls.__class_next__()
    except NotImplementedError:
      return MetaType.__next__(cls)  # NOQA
    else:
      return out

  def __bool__(cls) -> bool:
    """
    Here, too, I saw a nation of lost souls,
    far more than were above: they strained their chests
    against enormous weights, and with mad howls
    rolled them at one another. Then in haste
    they rolled them back, one party shouting out:
    "Why do you hoard?" and the other: "Why do you waste?"
    """
    try:
      out = cls.__class_bool__()
    except NotImplementedError:
      try:
        out = len(cls)  # NOQA
      except Exception as exception:
        return True
      else:
        return True if out else False
    else:
      return True if out else False

  def __contains__(cls, item: Any) -> bool:
    try:
      out = cls.__class_contains__(item)
    except NotImplementedError:
      out = None
      try:
        for clsItem in cls:
          if clsItem == item:
            out = True
            break
        else:
          out = False
      except Exception as exception:
        infoSpec = """argument of type '%s' is not iterable"""
        info = infoSpec % type(cls).__name__
        raise TypeError(info) from exception
      else:
        return True if out else False
    else:
      return True if out else False

  def __len__(cls) -> int:
    try:
      out = cls.__class_len__()
    except NotImplementedError:
      try:
        countingWithAsger = sum([1 for derp in cls])
      except Exception as exception:  # NOQA
        return len(MetaType)  # NOQA
      else:
        return countingWithAsger
    else:
      return out

  def __hash__(cls) -> int:
    try:
      out = cls.__class_hash__()
    except NotImplementedError:
      baseNames = [b.__name__ for b in cls.__bases__]
      metaName = type(cls).__name__
      nameTuple = (cls.__name__, *baseNames, metaName)
      return hash(nameTuple)  # NOQA
    else:
      return out

  #  DO NOT REMOVE THE FOLLOWING COMMENTED OUT METHODS
  # def __eq__(cls, other: Any) -> bool:
  #   """This method is disabled because of highly undefined behaviour!"""
  #
  #  def __ne__(cls, other: Any) -> bool:
  #    """See above"""

  #  DO NOT REMOVE THIS COMMENTED OUT METHOD  (not related to above)
  # def __getitem__(cls, item: Any) -> Any:
  #   """
  #   This method is intentionally commented out — not removed — to ensure
  #   discoverability and traceability.
  #   """
  #
  #   For an explanation, see the '__class_getitem__' note in the 'NameHook'
  #   class, located in 'worktoy.mcls.space_hooks'. In short, the interpreter
  #   handles '__class_getitem__' directly as of Python 3.7, making this
  #   override unnecessary and potentially conflicting.

  def __setitem__(cls, item: Any, value: Any) -> None:
    try:
      out = cls.__class_setitem__(item, value)
    except NotImplementedError:
      return MetaType.__setitem__(cls, item, value)  # NOQA
    else:
      return None

  def __delitem__(cls, item: Any) -> None:
    try:
      out = cls.__class_delitem__(item)
    except NotImplementedError:
      return MetaType.__delitem__(cls, item)  # NOQA
    else:
      return None

  def __getattr__(cls, name: str) -> Any:
    """Do not use the 'dot' operator to access class attributes during
    this method! Instead, the clunky 'object.__getattribute__' must be
    used to avoid infinite recursion. This is not an edge case, if you
    bring the 'dot' operator into an implementation here, it is recursion
    time!"""
    try:
      classGetAttr = object.__getattribute__(cls, '__class_getattr__')
      out = classGetAttr.__func__(cls, name)
    except NotImplementedError:
      raise AttributeError(name)
    else:
      return out

  def __setattr__(cls, name: str, value: Any) -> None:
    try:
      out = cls.__class_setattr__(name, value)
    except NotImplementedError:
      return MetaType.__setattr__(cls, name, value)  # NOQA
    else:
      return None

  def __delattr__(cls, name: str) -> None:
    try:
      out = cls.__class_delattr__(name)
    except NotImplementedError:
      return MetaType.__delattr__(cls, name)  # NOQA
    else:
      return None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @staticmethod
  def _notifySubclassHook(cls, *bases) -> type:
    """The _notifySubclassHook method is invoked to notify each baseclass
    of the created class of the class creation."""
    for base in bases:
      hook = getattr(base, '__subclasshook__', None)
      hook(cls)
    return cls

  def getNamespace(cls) -> ASpace:
    """Get the namespace object for the class."""
    return getattr(cls, '__namespace__', )



"""
LOCKED provides a sentinel object to indicate that an attribute may not be
retrieved in the current state of the object. Similar to 'DELETED',
but raises a 'RuntimeError', or subclass thereof, on access attempts.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import Sentinel


class LOCKED(Sentinel):
  """
  LOCKED is a sentinel object that indicates an attribute is locked and
  cannot be accessed in the current state of the object.
  """
  pass



"""Function objects decorated with the @overload decorator may have same
name but different signatures. The overload decorator is used to"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import TypeSig

if TYPE_CHECKING:  # pragma: no cover
  from typing import Callable


def overload(*types: object, **kwargs: object) -> Callable:
  """Function objects decorated with the @overload decorator may have same
  name but different signatures. The overload decorator is used to
  create a function object that can be called with different argument
  types. """

  typeSig = TypeSig(*types)

  def hereIsMyNumber(callMeMaybe: Callable) -> Callable:
    """Here is my number"""
    existing = getattr(callMeMaybe, '__type_sigs__', ())
    setattr(callMeMaybe, '__type_sigs__', (*[*existing, typeSig],))
    setattr(callMeMaybe, '__is_overloaded__', True)

    return callMeMaybe

  return hereIsMyNumber



"""
PreClass provides a stateful class containing the name and hash of a class
about to be created.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..waitaminute import MissingVariable, TypeException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class PreClass(type):
  """PreClass provides a stateful class containing the name and hash of a
  class about to be created."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __private_fallback__ = '__pre_class__'

  #  Private variables
  __meta_class__ = None
  __hash_value__ = None

  #  Public variables

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __new__(mcls, *args, **kwargs) -> type:
    """
    PreClass instances are basically treated as classes, thus 'type' is
    a reasonable base.
    """
    _name, _hash, bases, _meta = None, None, [], None
    for arg in args:
      if isinstance(arg, str):
        _name = arg
        continue
      if isinstance(arg, int):
        _hash = arg
        continue
      if isinstance(arg, type):
        _meta = arg
        continue
      if isinstance(arg, (tuple, list)):
        bases = (*arg,)
        continue
      raise TypeException('arg', arg, str, int, type, tuple, list)

    cls = type.__new__(mcls, _name, (*bases,), {})
    setattr(cls, '__hash_value__', _hash)
    setattr(cls, '__meta_class__', _meta)
    return cls

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instancecheck__(cls, instance: Any) -> bool:
    """
    Checks if the instance is an instance of the PreClass.
    """
    return True if hash(cls) == hash(instance) else False

  def __hash__(cls, ) -> int:
    """
    Returns the explicitly set hash value of the PreClass object.
    """
    if cls.__hash_value__ is None:
      raise MissingVariable('__hash_value__', int)
    if isinstance(cls.__hash_value__, int):
      return cls.__hash_value__
    name, value = '__hash_value__', cls.__hash_value__
    raise TypeException(name, value, int)

  def __getattribute__(cls, key: str, ) -> Any:
    """
    This reimplementation of __getattribute__ was done by a highly skilled
    professional, do not try this at home!
    """
    if key == '__class__':
      return object.__getattribute__(cls, '__meta_class__')
    return object.__getattribute__(cls, key)



"""
The 'worktoy.waitaminute.desc' package provides the custom exceptions
specific to the implementation of the descriptor protocol in the 'worktoy'
library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._descriptor_exception import DescriptorException
from ._access_error import AccessError
from ._protected_error import ProtectedError
from ._read_only_error import ReadOnlyError
from ._without_exception import WithoutException

__all__ = [
    'DescriptorException',
    'AccessError',
    'ReadOnlyError',
    'ProtectedError',
    'WithoutException',
]



"""
The 'worktoy.core' module provides the most primitive objects used by the
'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import sentinels
from ._instance import ContextInstance
from ._meta_type import MetaType
from ._meta_object import MetaObject
from ._object import Object

__all__ = [
    'sentinels',
    'ContextInstance',
    'MetaType',
    'MetaObject',
    'Object',
]



"""
The 'stringList' takes a string and returns a list of strings containing
substrings separated by a separator (default: ', ').

Keyword argument 'sep' specifies the separator(s) to use. More than one
separator may be specified by passing a non-empty iterable of strings.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import TypeException


def stringList(*args: str, **kwargs) -> list[str]:
  """
  Splits input strings using all provided separators in sequence.

  Keyword-only argument 'sep' may be a string or list of strings. Each
  separator is applied in turn, so that later separators act on results
  of earlier splits.
  """
  sep = kwargs.get('separator', ', ')
  if isinstance(sep, str):
    out = []
    for arg in args:
      out.extend(arg.split(sep))
    return [a.strip() for a in out if a.strip()]
  if isinstance(sep, (list, tuple)):
    out = [*args, ]
    for s in sep:
      out = stringList(*out, separator=s)
    return [a.strip() for a in out if a.strip()]
  raise TypeException('separator', sep, str, list, tuple)



"""Examples"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from .type_sig_example import main as typeSigExample

__all__ = [
    "typeSigExample",
]



"""
The 'worktoy.desc' module provides the base descriptor classes. This
module introduces a novel concept: descriptor-context.

When a descriptor is accessed through the owning class, the descriptor
object itself returns. When through an instance, the descriptor usually
performs the relevant accessor function as appropriate for the instance
received. This module expands this concept by introducing the
descriptor-context.

Methods decorated with the '@WITH' decorator require an active context.
These methods should have the following signature:

@WITH
def foo(self, instance: Any, owner: Type[type], **kwargs) -> Any:
...

Descriptor classes that are not stateless, must reimplement the
'exitContext' such that it restores the descriptor instance to its
previous state. Stateless descriptor classes need not reimplement the
method inherited from the 'Object' class.



"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._alias import Alias
from ._field import Field
from ._attri_box import AttriBox

__all__ = [
    'Alias',
    'Field',
    'AttriBox',
]



"""
MetaclassException provides a vastly improved exception to replace the
default 'all your base are belong to us' exception raised by metaclass
incompatibility.

For example, let MetaFoo and MetaBar both be metaclasses based on type.
Let Bar be derived from MetaBar. Then when trying to derive Foo from
MetaFoo whilst also based on Bar, we have the incompatibility:

class MetaFoo(type): pass

class MetaBar(type): pass

class Bar(metaclass=MetaBar): pass

class Foo(Bar, metaclass=MetaFoo): pass  # raises

When trying to create the class Foo, we get this bit of syntactic broccoli:
'TypeError: metaclass conflict: the metaclass of a derived class must be a
(non-strict) subclass of the metaclasses of all its bases'

What this is trying to communicate is that baseclasses must be derived
from the same metaclass or subclass as the new class. MetaclassException
will instead provide a much clearer message that also references
specifically which base class and metaclass is causing the issue.

Metaclass conflict while defining class 'Foo':

- Foo derives from metaclass 'MetaFoo'
- While also based on class 'Bar' derived from 'MetaBar'
- But 'MetaBar' is not 'MetaFoo' nor a subclass of 'MetaFoo'.

Classes must be based on baseclasses derived from the same metaclass or
subclass of it.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class MetaclassException(TypeError):
  """
  Vastly improved exception for metaclass incompatibility.
  """

  __slots__ = ('name', 'meta', 'badBase', 'badMeta')

  def __init__(self, mcls: type, name: str, *bases: type) -> None:
    """Initialize the MetaclassException object."""
    self.name = name
    self.meta = mcls
    for base in bases:
      if isinstance(base, mcls):
        continue
      self.badBase = base
      self.badMeta = type(base)
      break
    else:
      infoSpec = """
      MetaclassException was raised with no base classes incompatible 
      with the received metaclass '%s' while defining class '%s'."""
      info = infoSpec % (mcls.__name__, name)
      raise TypeError(textFmt(info))
    TypeError.__init__(self, )

  def __str__(self) -> str:
    """
    Return a string representation of the MetaclassException object.
    """
    infoSpec = """Metaclass conflict while deriving class: '%s' from 
    metaclass: '%s', because the base class: '%s' is derived from the 
    different metaclass: '%s' which is not a subclass of '%s'."""

    names = [
        self.name,
        self.meta.__name__,
        self.badBase.__name__,
        self.badMeta.__name__,
        self.meta.__name__
    ]

    info = infoSpec % (*names,)
    return textFmt(info)

  __repr__ = __str__



"""
The 'factorial' function computes the factorial of a non-negative integer.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations


def factorial(n: int, c: int = 1) -> int:
  if n < -1:
    infoSpec = """The implementation of factorial in 
    'worktoy.utilities.mathematics' does not implement the full gamma 
    function. Try using the 'scipy.special.gamma' function instead."""
    from .. import textFmt
    raise ValueError(textFmt(infoSpec))
  if n in [0, 1]:
    return c
  return factorial(n - 1, c * n)



"""ReservedName is raised to indicate that a used name is reserved."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class ReservedName(Exception):
  """ReservedName is raised to indicate that a used name is reserved."""

  __slots__ = ('resName',)

  def __init__(self, name: str) -> None:
    """Initialize the ReservedName object."""
    self.resName = name
    info = """Attempted to use reserved name: '%s'!"""
    Exception.__init__(self, info % name)

  def __str__(self) -> str:
    """
    String representation of the ReservedName exception.
    """
    infoSpec = """Attempted to use reserved name: '%s'!"""
    info = infoSpec % self.resName
    return textFmt(info)

  __repr__ = __str__



"""
The 'worktoy.waitaminute.meta' provides the custom exceptions used by the
class creation flow in 'worktoy.core'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._illegal_instantiation import IllegalInstantiation
from ._hook_exception import HookException
from ._duplicate_hook import DuplicateHook
from ._metaclass_exception import MetaclassException
from ._del_exception import DelException
from ._questionable_syntax import QuestionableSyntax
from ._reserved_name import ReservedName

__all__ = [
    'DuplicateHook',
    'IllegalInstantiation',
    'HookException',
    'MetaclassException',
    'DelException',
    'QuestionableSyntax',
    'ReservedName',
]



"""
AbstractSpaceHook provides an abstract baseclass for hooks used by the
namespaces in the metaclass system.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import SpaceDesc
from ...core import Object

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Callable, TypeAlias, Type
  from worktoy.mcls import AbstractNamespace as ASpace

  AccessorHook = Callable[[ASpace, str, Any], Any]
  CompileHook = Callable[[ASpace, dict], dict]
  Space: TypeAlias = Type[ASpace]
  Meta: TypeAlias = Type[type]


class AbstractSpaceHook(Object):
  """
  AbstractSpaceHook is the abstract base class for defining hook objects
  used in conjunction with AbstractNamespace. These hooks enable modular,
  stage-specific interception during class body evaluation and namespace
  compilation within the metaclass system.

  ## Purpose

  Hooks allow custom behavior to be injected into the class construction
  pipeline without modifying the namespace or metaclass core logic. They
  are used to observe and/or alter how names are accessed, assigned, or
  compiled into the final class definition.

  ## Integration

  To activate a hook, simply instantiate a subclass of AbstractSpaceHook
  inside the body of a namespace class (i.e., a subclass of
  AbstractNamespace). The descriptor protocol (`__set_name__`) ensures the
  hook registers itself with the namespace automatically at definition time.

  Example:

      class MyNamespace(AbstractNamespace):
        overloadHook = OverloadPhase()
        validationHook = ReservedNamePhase()

  ## Lifecycle Hook Methods

  Subclasses may override any of the following methods to participate in
  different stages of the namespace lifecycle. All are optional.

  - `setItemPhase(self, key, value, oldValue) -> bool`
    Called just before a name is set in the namespace.
    Returning True blocks the default behavior.

  - `getItemPhase(self, key, value) -> bool`
    Called just before a name is retrieved from the namespace.
    Returning True blocks the default behavior.

  - `preCompilePhase(self, compiled: dict) -> dict`
    Called after the class body finishes executing, but before the
    namespace is finalized. May transform or replace namespace contents.

  - `postCompilePhase(self, compiled: dict) -> dict`
    Called immediately before the finalized namespace is handed off to the
    metaclass. Can be used for final transformations or validation.

  ## Descriptor Behavior

  AbstractSpaceHook implements the descriptor protocol. When accessed via a
  namespace class, it is bound with the following attributes:

  - `self.space` refers to the active namespace instance.
  - `self.spaceClass` refers to the namespace class itself.

  These attributes can be used to introspect the environment the hook is
  participating in.

  ## Extension Notes

  Subclasses are expected to override only the relevant hook methods.
  If none are overridden, the hook has no effect.

  The `addPhase()` method of the namespace class is automatically invoked
  during registration. Hook authors do not need to call it manually.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __space_object__ = None  # The owning namespace instance

  #  Public variables
  space = SpaceDesc()  # The namespace instance this hook is bound to

  #  TYPE_CHECKING
  if TYPE_CHECKING:  # pragma: no cover
    from . import AbstractSpaceHook
    from .. import AbstractNamespace
    assert isinstance(AbstractSpaceHook.space, AbstractNamespace)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  def getItemPhase(self, key: str, value: Any, ) -> bool:
    """Hook for getItem. This is called before the __getitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """

  def setItemPhase(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """

  def preCompilePhase(self, compiledSpace: dict) -> dict:
    """Hook for preCompile. This is called before the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    return compiledSpace

  def postCompilePhase(self, compiledSpace: dict) -> dict:
    """Hook for postCompile. This is called after the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    return compiledSpace

  def newClassPhase(self, cls: Meta, ) -> Meta:
    """
    Final phase invoked by the metaclass after it has created the new
    class object, but before returning it. This phase occurs before the
    normal post class creation flow continues.
    """
    return cls

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: Space, name: str, **kwargs) -> None:
    """
    After the super call, adds one self to the namespace class as a hook
    class.
    """
    super().__set_name__(owner, name, )
    owner.addHook(self)

  def __get__(self, instance: ASpace, owner: Space, **kwargs) -> Any:
    """
    Descriptor protocol method. Returns the bound hook instance with
    `space` and `spaceClass` attributes set to the current namespace
    instance and its class.
    """
    self.__space_object__ = instance
    return self



"""FidGen provides filename generator. Given a format specification and a
directory, it returns the next available filename of the given format."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from ..desc import Field
from ..mcls import BaseObject
from ..utilities import maybe, stringList
from ..static import overload
from ..waitaminute import TypeException
from . import validateExistingDirectory, validateAvailablePath

if TYPE_CHECKING:  # pragma: no cover
  from typing import Optional, Any, TypeAlias

  ArgRes: TypeAlias = tuple[Optional[str], tuple[Any, ...]]


class FidGen(BaseObject):
  """
  FidGen provides filename generator. Given a format specification and a
  directory, it returns the next available filename of the given format.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __fallback_extension__ = 'json'
  __fallback_directory__ = os.path.abspath(os.path.dirname(__file__))

  #  Private variables
  __base_name__ = None
  __generated_names__ = None
  __file_spec__ = None
  __file_extension__ = None
  __file_directory__ = None

  #  Public variables
  fileExtension = Field()
  fileDirectory = Field()
  fileSpec = Field()

  #  Virtual variables
  filePath = Field()
  nextName = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getGeneratedNames(self) -> list[str]:
    """
    Getter-function for the generated names.
    """
    return maybe(self.__generated_names__, [])

  def _createFileSpec(self) -> None:
    """
    Creator-function for the file specification.
    """
    baseName = maybe(self.__base_name__, type(self).__name__, )
    chars = []
    for i, char in enumerate(baseName):
      if char.upper() == char:
        if i:
          chars.append('_%s' % char.lower())
          continue
        chars.append(char.lower())
        continue
      chars.append(char)
    snakeName = ''.join(chars)
    spec = """%s%%03d.%s""" % (snakeName, self.fileExtension)
    self.__file_spec__ = spec

  @fileSpec.GET
  def _getFileSpec(self, **kwargs) -> str:
    """Get the file specification."""
    if self.__file_spec__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._createFileSpec()
      return self._getFileSpec(_recursion=True, )
    if isinstance(self.__file_spec__, str):
      return str(os.path.join(self.fileDirectory, self.__file_spec__))
    raise TypeException('__file_spec__', self.__file_spec__, str)

  @fileExtension.GET
  def _getFileExtension(self, **kwargs) -> str:
    """Get the file extension."""
    return maybe(self.__file_extension__, self.__fallback_extension__, )

  @fileDirectory.GET
  def _getFileDirectory(self, **kwargs) -> str:
    """Get the file directory."""
    return maybe(self.__file_directory__, self.__fallback_directory__, )

  @nextName.GET
  def _getNextName(self, n=None) -> str:
    """
    Retrieves the next available filename based on the file specification.
    """
    n = maybe(n, 0)  # easier to cover wink wink
    while not validateAvailablePath(self.fileSpec % n, strict=False):
      n += 1
      if n > 100:
        break
    else:
      out = self.fileSpec % n
      self._addGeneratedName(out)
      return out
    raise RecursionError

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _addGeneratedName(self, name: str) -> None:
    """
    Adds a generated name to the list of generated names.
    """
    existing = self._getGeneratedNames()
    self.__generated_names__ = [*existing, name]

  @fileExtension.SET
  def _setFileExtension(self, value: str) -> None:
    """Set the file extension."""
    if not isinstance(value, str):
      raise TypeException('__file_extension__', value, str)
    if not value:
      raise ValueError('__file_extension__ must be a non-empty string')
    self.__file_extension__ = value

  @fileDirectory.SET
  def _setFileDirectory(self, value: str, **kwargs) -> None:
    """Set the file directory."""
    self.__file_directory__ = validateExistingDirectory(value)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @overload(str)
  def __init__(self, fileName: str, **kwargs) -> None:
    """Initialize the FidGen object."""
    self.__base_name__ = fileName
    if kwargs:
      self.__init__(**kwargs)

  @overload(str, str, str)
  @overload(str, str)
  def __init__(self, *args, **kwargs) -> None:
    posArgs = (*args,)
    argDir, posArgs = self._findDirectory(*posArgs)
    argExt, posArgs = self._findFileExtension(*posArgs)
    for arg in posArgs:
      self.baseName = arg
    if argDir is not None:
      self.fileDirectory = argDir
    if argExt is not None:
      self.fileExtension = argExt
    if kwargs:
      self.__init__(**kwargs)

  @overload()  # kwargs
  def __init__(self, **kwargs) -> None:
    """Initialize the FidGen object."""
    nameKeys = stringList("""name, file, fileName, filename, file_name""")
    extKeys = stringList("""ext, extension, file_extension""")
    dirKeys = stringList("""dir, directory, file_directory""")
    name, kwargs = self.parseKwargs(str, *nameKeys, **kwargs)
    ext, kwargs = self.parseKwargs(str, *extKeys, **kwargs)
    dir_, kwargs = self.parseKwargs(str, *dirKeys, **kwargs)
    if all([i is not None for i in [name, ext, dir_]]):
      self.__init__(name, ext, dir_, )

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type, **kwargs) -> FidGen:
    """
    When the owning class is created, Python calls this method to allowing
    the type signatures to be updated with the owner class. This is
    necessary as the type signatures are able to reference the owning
    class before it is created by using the 'THIS' token object in place
    of it.
    """
    self.__context_caller__ = instance
    return self.__instance_get__()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @staticmethod
  def _findDirectory(*args) -> ArgRes:
    """Finds the directory in positional arguments. """
    unusedArgs = []
    posArgs = [*reversed(args), ]
    while posArgs:
      arg = posArgs.pop()
      if isinstance(arg, str):
        if os.path.isdir(arg):
          out = arg
          unusedArgs.extend(posArgs)
          break
        unusedArgs.append(arg)
    else:
      return None, (*args,)
    return out, (*unusedArgs,)

  @staticmethod
  def _getCommonExtensions() -> list[str]:
    """Returns a list of common file extensions."""
    return stringList(
        """json, txt, csv, xml, html, pdf, doc, csv, py, 
        mkv, mp4, mp3, wav, jpg, png, gif, zip, tar, gz, bz2"""
    )

  @classmethod
  def _findFileExtension(cls, *args) -> ArgRes:
    """Finds the file extension in positional arguments. """
    unusedArgs = []
    posArgs = [*reversed(args), ]
    while posArgs:
      arg = posArgs.pop()
      if isinstance(arg, str):
        if str.startswith(arg, '*.'):
          out = arg[2:]
          unusedArgs.extend(posArgs)
          break
        if arg in cls._getCommonExtensions():
          out = arg
          unusedArgs.extend(posArgs)
          break
        unusedArgs.append(arg)
    else:
      return None, (*args,)
    return out, (*unusedArgs,)



"""
The 'yeetDirectory' function removes a directory with all contents.
Effectively the same as: 'rm -rf <directory>'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from ..utilities import textFmt
from ..waitaminute import PathSyntaxException

if TYPE_CHECKING:  # pragma: no cover
  pass


def yeetDirectory(dirPath: str, **kwargs) -> None:
  """
  Removes a directory with all contents. Effectively the same as: 'rm -rf
  <directory>'.

  Args:
    dirPath (str): The path to the directory to remove.

  Raises:
    FileNotFoundError: If the directory does not exist.
    NotADirectoryError: If the path is not a directory.
    PathSyntaxException: If the path is not absolute.
  """
  if not os.path.isabs(dirPath):
    raise PathSyntaxException(dirPath)
  if not os.path.exists(dirPath):
    if not kwargs.get('strict', True):
      return
    infoSpec = """No directory exists at: '%s'!"""
    info = textFmt(infoSpec % dirPath)
    raise FileNotFoundError(info)
  for item in os.listdir(dirPath):
    itemPath = os.path.join(dirPath, item)
    if os.path.isdir(itemPath):
      yeetDirectory(itemPath, strict=False)
      continue
    os.remove(itemPath)
  else:
    os.rmdir(dirPath)



"""This files provides common type names used by the mcls package. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from worktoy.mcls import AbstractNamespace

  Space = AbstractNamespace
  Spaces = tuple[AbstractNamespace, ...]
  Base = tuple[type, ...]
  Types = tuple[type, ...]

else:
  Space = object
  Spaces = tuple
  Base = tuple
  Types = tuple



"""
DuplicateKeeNum is a custom exception raised to indicate that a KeeNum
class received a duplicate entry for an enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class DuplicateKeeNum(Exception):
  """
  DuplicateKeeNum is a custom exception raised to indicate that a KeeNum
  class received a duplicate entry for an enumeration.
  """

  __slots__ = ('memberName', 'memberValue')

  def __init__(self, name: str, value: Any) -> None:
    """Initialize the DuplicateKeeNum object."""
    self.memberName = name
    self.memberValue = value
    Exception.__init__(self, )

  def __str__(self, ) -> str:
    """Return the string representation of the DuplicateKeeNum object."""
    infoSpec = """Duplicate KeeNum member '%s' with value '%s'!"""
    info = infoSpec % (self.memberName, str(self.memberValue))
    return textFmt(info)

  __repr__ = __str__



"""The 'worktoy.waitaminute.keenum' module provides custom exceptions used
by the 'worktoy.keenum' module."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._duplicate_kee_num import DuplicateKeeNum
from ._empty_kee_num_error import EmptyKeeNumError
from ._kee_num_type_exception import KeeNumTypeException

__all__ = [
    'DuplicateKeeNum',
    'EmptyKeeNumError',
    'KeeNumTypeException',
]



"""
KeeNumTypeException is a custom exception class raised to indicate a
member of a KeeNum enumeration with inconsistent value type.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeNumTypeException(TypeError):
  """
  KeeNumTypeException is a custom exception class raised to indicate a
  member of a KeeNum enumeration with inconsistent value type.
  """

  __slots__ = ('memberName', 'memberValue', 'expectedType')

  def __init__(self, name: str, value: Any, expectedType: type) -> None:
    """Initialize the KeeNumTypeException object."""
    self.memberName = name
    self.memberValue = value
    self.expectedType = expectedType
    TypeError.__init__(self, )

  def __str__(self, ) -> str:
    """Return the string representation of the KeeNumTypeException object."""
    infoSpec = """KeeNum member '%s' has value '%s' of type '%s', but 
    expected type is '%s'!"""
    name = self.memberName
    value = str(self.memberValue)
    valueType = type(self.memberValue).__name__
    expType = self.expectedType.__name__
    info = infoSpec % (name, value, valueType, expType)
    return textFmt(info)

  __repr__ = __str__



"""VariableNotNone should be raised when a variable is unexpectedly not
None."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class VariableNotNone(Exception):
  """VariableNotNone should be raised when a variable is unexpectedly not
  None."""

  __slots__ = ('name', 'value')

  def __init__(self, *args) -> None:
    """Initialize the VariableNotNone object."""
    self.name, self.value, *_ = [*args, None, None]
    Exception.__init__(self, )

  def __str__(self, ) -> str:
    """Get the info spec."""
    infoSpec = """Unexpected value: '%s' at name '%s' expected to be 
    None!"""
    valueStr = textFmt(str(self.value), )
    name = self.name
    return textFmt(infoSpec % (valueStr, name), )

  __repr__ = __str__



"""
KeeHook provides the namespace hook for the KeeSpace namespace class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls.space_hooks import AbstractSpaceHook
from . import _AutoMember
from ..waitaminute.keenum import KeeNumTypeException, DuplicateKeeNum

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeSpaceHook(AbstractSpaceHook):
  """
  KeeHook provides the namespace hook for the KeeSpace namespace class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _addMember(self, name: str, value: Any = None, **kwargs) -> None:
    """
    Add a member to the KeeSpace namespace.
    """
    if isinstance(value, _AutoMember):
      value = value.getValue()
    if value is None:
      if kwargs.get('_recursion2', False):
        raise RecursionError  # pragma: no cover
      return self._addMember(name, name, _recursion2=True)
    valueType = self._getValueType()
    if valueType is None:
      if kwargs.get('_recursion', False):
        raise RecursionError  # pragma: no cover
      self._setValueType(type(value))
      return self._addMember(name, value, _recursion=True)
    else:
      if not isinstance(value, valueType):
        raise KeeNumTypeException(name, value, valueType)
    existing = self.space.get('__future_entries__', {})
    if name in existing:
      raise DuplicateKeeNum(name, value)
    existing[name] = value
    self.space['__future_entries__'] = existing

  def _setValueType(self, valueType: type, ) -> None:
    """
    Set the value type for the KeeSpace namespace.
    """
    self.space['__future_value_type__'] = valueType

  def _getValueType(self) -> type:
    """
    Get the value type for the KeeSpace namespace.
    """
    return self.space.get('__future_value_type__', None)

  def setItemPhase(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """
    The setItemHook method is called when an item is set in the
    namespace.
    """
    if key.startswith('__') and key.endswith('__'):
      # Skip special keys
      return False
    if callable(value):
      return False
    self._addMember(key.upper(), value)
    return False

  def preCompilePhase(self, compiledSpace: dict) -> dict:
    """Hook for preCompile. This is called before the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    compiledSpace['__future_entries__'] = dict()
    return compiledSpace



"""
The 'worktoy.mcls.space_hooks' package provides the hooks used by the
Namespace system.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._space_desc import SpaceDesc
from ._reserved_names import ReservedNames
from ._abstract_space_hook import AbstractSpaceHook
from ._reserved_namespace_hook import ReservedNamespaceHook
from ._pre_class_hook import PreClassSpaceHook
from ._overload_hook import OverloadSpaceHook
from ._name_hook import NamespaceHook

__all__ = [
    'SpaceDesc',
    'ReservedNames',
    'AbstractSpaceHook',
    'ReservedNamespaceHook',
    'PreClassSpaceHook',
    'OverloadSpaceHook',
    'NamespaceHook',
]



"""The 'worktoy.workWrap' function receives an integer defining character
width and any number of strings. The function then returns a list of
strings containing the words from the strings received such that each
entry in the list does not exceed the character width. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations


def wordWrap(width: int, *textLines, **kwargs) -> str:
  """The wordwrap function wraps the input text to a specified width."""
  newLine = kwargs.get('newLine', '<br>').strip().lower()
  if not isinstance(width, int):
    from ..waitaminute import TypeException
    raise TypeException('width', width, int)
  words = []
  for line in textLines:
    if not isinstance(line, str):
      from ..waitaminute import TypeException
      raise TypeException('line', line, str)
    words.extend(line.split())
  lines = []
  line = []
  while words:
    word = words.pop(0)
    if word.lower() == newLine or len(' '.join([*line, word])) > width:
      lines.append(' '.join(line))
      line = []
      continue
    line.append(word)
  return '\n'.join(lines)



"""INSTANCE provides the contextual instance of an 'Object' object."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class ContextInstance:
  """INSTANCE provides the contextual instance of an 'Object' object."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type) -> Any:
    """Returns the instance of the 'Object' instance."""
    if instance is None:
      return self
    return instance.getContextInstance()



"""
The 'maybe' function returns the first value that is not 'None'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


def maybe(*args) -> Any:
  """
  Returns the first argument that is not None.
  """
  for arg in args:
    if arg is not None:
      return arg



"""EZHook collects the field entries in EZData class bodies. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType
from typing import TYPE_CHECKING

from ..mcls.space_hooks import AbstractSpaceHook, ReservedNames
from ..waitaminute.meta import ReservedName

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Iterator
  from worktoy.ezdata import EZData


class EZSpaceHook(AbstractSpaceHook):
  """EZHook collects the field entries in EZData class bodies. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __auto_names__ = """__slots__, __init__, __eq__, __str__, __repr__,
    __iter__, __getitem__, __setitem__, __getattr__"""

  #  Public Variables
  reservedNames = ReservedNames()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Hook methods

  def preCompilePhase(self, compiledSpace: dict) -> dict:
    """The preCompileHook method is called before the class is compiled."""
    dataFields = self.space.getDataFields()
    compiledSpace['__slots__'] = (
        *[dataField.key for dataField in dataFields],)
    return compiledSpace

  def postCompilePhase(self, compiledSpace: dict) -> dict:
    """The postCompileHook method is called after the class is compiled."""
    dataFields = self.space.getDataFields()
    initMethod = self.initFactory(*dataFields)
    eqMethod = self.eqFactory(*dataFields)
    hashMethod = self.hashFactory(*dataFields)
    strMethod = self.strFactory(*dataFields)
    reprMethod = self.reprFactory(*dataFields)
    iterMethod = self.iterFactory(*dataFields)
    lenMethod = self.lenFactory(*dataFields)
    getItemMethod = self.getItemFactory(*dataFields)
    setItemMethod = self.setItemFactory(*dataFields)
    compiledSpace['__init__'] = initMethod
    compiledSpace['__eq__'] = eqMethod
    compiledSpace['__hash__'] = hashMethod
    compiledSpace['__str__'] = strMethod
    compiledSpace['__repr__'] = reprMethod
    compiledSpace['__iter__'] = iterMethod
    compiledSpace['__len__'] = lenMethod
    compiledSpace['__getitem__'] = getItemMethod
    compiledSpace['__setitem__'] = setItemMethod
    return compiledSpace

  def setItemPhase(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """The setItemHook method is called when an item is set in the
    enumeration."""
    if key in self.__auto_names__:
      if not hasattr(value, '__is_root__'):
        raise ReservedName(key)
    if key in self.reservedNames:
      return False  # Already handled by ReservedNameHook
    if callable(value):
      return False
    self.space.addField(key, type(value), value)
    return True

  # \_____________________________________________________________________/ #
  #  Method factories
  # /¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨\ #
  @staticmethod
  def initFactory(*dataFields) -> FunctionType:
    """
    Creates the '__init__' method for the EZData class.
    """

    slotKeys = [dataField.key for dataField in dataFields]
    defVals = [dataField.val for dataField in dataFields]

    def __init__(self, *args, **kwargs) -> None:
      """
      The generated '__init__' method sets attributes on the instance
      based on given arguments. Keyword arguments take precedence.
      Positional arguments are applied in order.
      """
      posArgs = [*args, ]
      while len(posArgs) < len(slotKeys):
        posArgs.append(None)
      for key, defVal in zip(slotKeys, defVals):
        setattr(self, key, defVal)
      for key, arg in zip(slotKeys, posArgs):
        if arg is not None:
          setattr(self, key, arg)
      for key in slotKeys:
        if key in kwargs:
          setattr(self, key, kwargs[key])

    return __init__

  @staticmethod
  def eqFactory(*dataFields) -> FunctionType:
    """
    Creates the '__eq__' method for the EZData class.
    """

    def __eq__(self, other: EZData) -> bool:
      """
      Instances of EZData are equal if each of their data fields are equal.
      """
      if type(self) is not type(other):
        return NotImplemented
      for dataField in dataFields:
        key = dataField.key
        if getattr(self, key) != getattr(other, key):
          return False
      return True

    return __eq__

  @staticmethod
  def hashFactory(*dataFields) -> FunctionType:
    """
    Creates the '__hash__' method for the EZData class.
    """

    def __hash__(self) -> int:
      """
      The hash of an EZData instance is the hash of its data fields.
      """
      hashVal = 0
      for dataField in dataFields:
        key = dataField.key
        val = getattr(self, key)
        hashVal ^= hash(val)
      return hashVal

    return __hash__

  @staticmethod
  def strFactory(*dataFields) -> FunctionType:
    """The strFactory method is called when the class is created."""

    def __str__(self) -> str:
      """The __str__ method is called when the class is created."""
      clsName = type(self).__name__
      keys = [dataField.key for dataField in dataFields]
      vals = [str(getattr(self, key)) for key in keys]
      keyVals = ['%s=%s' % (key, val) for key, val in zip(keys, vals)]
      return """%s(%s)""" % (clsName, ', '.join(keyVals))

    return __str__

  @staticmethod
  def reprFactory(*dataFields) -> FunctionType:
    """The reprFactory method is called when the class is created."""

    def __repr__(self) -> str:
      """The __repr__ method is called when the class is created."""
      clsName = type(self).__name__
      keys = [dataField.key for dataField in dataFields]
      vals = [str(getattr(self, key)) for key in keys]
      return """%s(%s)""" % (clsName, ', '.join(vals))

    return __repr__

  @staticmethod
  def iterFactory(*dataFields) -> FunctionType:
    """The iterFactory method is called when the class is created."""

    def __iter__(self, ) -> Iterator:
      """
      Implementation of the iteration protocol
      """
      for key in self.__slots__:
        yield getattr(self, key)

    return __iter__

  @staticmethod
  def lenFactory(*dataFields) -> FunctionType:
    """The lenFactory method is called when the class is created."""

    def __len__(self) -> int:
      """The __len__ method is called when the class is created."""
      return len(self.__slots__)

    return __len__

  @staticmethod
  def getItemFactory(*dataFields) -> FunctionType:
    """The getItemFactory method is called when the class is created."""

    def __getitem__(self, key: str) -> object:
      """The __getitem__ method is called when the class is created."""
      if key in self.__slots__:
        return getattr(self, key)
      raise KeyError(key)

    return __getitem__

  @staticmethod
  def setItemFactory(*dataFields) -> FunctionType:
    """The setItemFactory method is called when the class is created."""

    def __setitem__(self, key: str, value: object) -> None:
      """The __setitem__ method is called when the class is created."""
      if key in self.__slots__:
        return setattr(self, key, value)
      raise KeyError(key)

    return __setitem__



"""DispatchException provides a custom exception raised when an instance
of OverloadDispatcher fails to resolve the correct function from the
given arguments. Because the overload protocol relies on type matching,
this exception subclasses TypeError such that it can be caught by external
error handlers. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from worktoy.static import Dispatch


class DispatchException(TypeError):
  """DispatchException provides a custom exception raised when an instance
  of OverloadDispatcher fails to resolve the correct function from the
  given arguments. Because the overload protocol relies on type matching,
  this exception subclasses TypeError such that it can be caught by external
  error handlers. """

  __slots__ = ('dispatchObject', 'receivedArguments')

  def __init__(self, dispatch: Dispatch, *args) -> None:
    self.dispatchObject = dispatch
    self.receivedArguments = args
    TypeError.__init__(self, )

  def __str__(self) -> str:
    """
    Return a string representation of the DispatchException.
    """
    ownerName = self.dispatchObject.getFieldOwner().__name__
    fieldName = self.dispatchObject.getFieldName()
    clsName = type(self.dispatchObject).__name__
    header = '%s object at %s.%s' % (clsName, ownerName, fieldName)
    typeArgs = [type(arg).__name__ for arg in self.receivedArguments]
    argStr = '%s' % ', '.join(typeArgs)
    typeSigs = self.dispatchObject.getTypeSigs()
    typeSigStr = [str(sig) for sig in typeSigs]
    sigStr = '<br><tab>'.join(typeSigStr)

    infoSpec = """%s received arguments with signature: <br><tab>%s
    <br>which does not match any of the expected signatures:<br><tab>%s"""
    return textFmt(infoSpec % (header, argStr, sigStr))

  __repr__ = __str__



"""TypeException is a custom exception class for handling type related
errors. Specifically, this exception should NOT be raised if the object is
None instead of the expected type. This is because None indicates absense
rather than type mismatch. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


class TypeException(TypeError):
  """
  TypeException is a custom exception class for handling type related
  errors. Specifically, this exception should NOT be raised if the object is
  None instead of the expected type.
  """

  __slots__ = ('varName', 'actualObject', 'actualType', 'expectedTypes',)

  def __init__(self, name: str, obj: object, *types) -> None:
    """Initialize the TypeException with the name of the variable, the
    received object, and the expected types."""
    TypeError.__init__(self, )
    self.varName = name
    self.actualObject = obj
    self.actualType = type(obj)
    self.expectedTypes = types

  def __str__(self) -> str:
    """String representation of the TypeException."""
    from ..utilities import joinWords, textFmt
    infoSpec = """Expected object at name '%s' to be an instance of %s, 
    but received object: '%s' of type '%s'!"""
    typeNames = [t.__name__ for t in self.expectedTypes]
    typeStr = joinWords(*["""'%s'""" % name for name in typeNames], )
    objStr = repr(self.actualObject)
    if len(objStr) > 50:
      objStr = """%s...""" % objStr[:47]
    clsType = type(self.actualObject).__name__
    info = infoSpec % (self.varName, typeStr, objStr, clsType)
    return textFmt(info)

  __repr__ = __str__



"""
This files provides mathematical constants used by the
'worktoy.utilities.mathematics' module.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


def _arctanPi(z: float, shift: float = None, **kwargs) -> float:
  """Using Taylor series to compute arctan to compute pi."""
  if shift is None:
    shift = 1e12
  term = z
  result = z
  n = 1

  while abs(term) > kwargs.get('epsilon', 1e-32):
    term *= -z * z
    result += term / (2 * n + 1)
    n += 1
  else:
    return shift * result


def _pi() -> float:
  """Returns the value of pi."""
  out = 44 * _arctanPi(1 / 57)
  out += 7 * _arctanPi(1 / 239)
  out -= 12 * _arctanPi(1 / 682)
  return (4 * out + 96 * _arctanPi(1 / 12943)) * 1e-12


pi = _pi()


def _arctan(x: float) -> float:
  """Returns the arctangent of x using the Taylor series expansion."""
  return _arctanPi(x, shift=1, )


def _arcTanUnit(x: float) -> float:
  """
  Implements a Taylor series expansion for the arctangent function
  for values near 1.
  """

  result = pi / 4
  term = (x - 1) / (x + 1)
  w = (x - 1) / (x + 1)
  n = 1
  while abs(term) > 1e-32:
    result += term / (2 * n - 1)
    term *= -w ** 2
    n += 1
    if n > 100:  # pragma: no cover
      break
  else:
    return result
  raise RecursionError  # pragma: no cover


def arcTan(x: float) -> float:
  """
  Returns the arctangent of a complex number.
  Uses the Taylor series expansion for values near 1.
  """
  if abs(x) < 1e-16:
    return 0.0
  if abs(x - 1) < 1e-16:
    return pi / 4
  if abs(x - 1) < 0.5:
    return _arcTanUnit(x)
  if x < 0:
    return -arcTan(-x)
  if x > 1:
    return pi / 2 - arcTan(1 / x)
  return _arctan(x)


def atan2(y: float, x: float) -> float:
  """Returns the arctangent of y/x, handling the quadrant correctly."""
  if x > 0:
    return arcTan(y / x)
  if x < 0:
    if y >= 0:
      return arcTan(y / x) + pi
    else:
      return arcTan(y / x) - pi
  if y > 0:
    return pi / 2
  if y < 0:
    return -pi / 2
  return 0.0



"""
TypeCastException is a custom exception class raised by the 'typeCast'
function from the 'worktoy.static' module to indicate that a value could
not be cast to the target type.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class TypeCastException(TypeError):
  """
  TypeCastException is a custom exception class raised by the 'typeCast'
  function from the 'worktoy.static' module to indicate that a value could
  not be cast to the target type.
  """

  __slots__ = ('type_', 'arg')

  def __init__(self, type_: type, arg: Any) -> None:
    """Initialize the TypeCastException with the value and target type."""
    self.type_ = type_
    self.arg = arg
    TypeError.__init__(self, )

  def __str__(self) -> str:
    """Get the string representation of the exception."""
    infoSpec = """Unable to cast value '%s' to type '%s'!"""
    typeStr = self.type_.__name__
    argStr = str(self.arg)
    info = infoSpec % (argStr, typeStr)
    from ...utilities import textFmt
    return textFmt(info)

  __repr__ = __str__



"""
SpaceDesc provides a descriptor class for 'AbstractSpaceHook' objects,
exposing it to the namespace object.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class SpaceDesc:
  """
  SpaceDesc provides a descriptor class for 'AbstractSpaceHook' objects,
  exposing it to the namespace object.
  """

  def __get__(self, instance: Any, owner: type) -> Any:
    """Returns the space hook of the 'AbstractSpaceHook' instance."""
    if instance is None:
      return self
    return getattr(instance, '__space_object__')



"""
THIS is the sentinel object representing the class currently under
construction. Similar to the
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class THIS(Sentinel):
  """
  THIS is the sentinel object representing the class currently under
  construction. Similar to the 'self' keyword, it is used to refer to the
  class being defined.
  """



"""The 'worktoy' package provides a collection of utilities leveraging
advanced python features including custom metaclasses and the descriptor
protocol. The readme file included provides detailed documentation on the
included features. The modules provided depend on each other in
implementation, but can be used independently.

The package consists of thr following modules:
- 'parse': For low-level parsing.
- 'text': For working with text.
- 'waitaminute': Provides custom exception classes.
- 'static': For parsing of objects.
- 'mcls': Provides custom metaclasses.
- 'attr': Provides custom descriptors.
- 'ezdata': Provides the 'EZData' class for creating data classes.
- 'keenum': Provides the 'KeeNum' class for creating enums.
"""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from . import utilities
from . import core
from . import waitaminute
from . import desc
from . import static
from . import mcls
from . import keenum
from . import ezdata



"""WriteOnceError is a custom error class raised to indicate that a
variable was attempted to be written to more than once."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any

  from worktoy.core import Object


class WriteOnceError(TypeError):
  """WriteOnceError is a custom error class raised to indicate that a
  variable was attempted to be written to more than once."""

  __slots__ = ('desc', 'oldValue', 'newValue')

  def __init__(self, desc: Object, oldVal: Any, newVal: Any) -> None:
    self.desc = desc
    self.oldValue = oldVal
    self.newValue = newVal
    TypeError.__init__(self, )

  def __str__(self) -> str:
    infoSpec = """Attempted to overwrite write-once attribute at '%s.%s' 
    having existing value: '%s' with new value: '%s'!"""
    owner = getattr(self.desc, '__field_owner__', object())  # no __name__
    ownerName = getattr(owner, '__name__', 'Unknown')
    fieldName = getattr(self.desc, '__field_name__', 'Unknown')
    oldStr = str(self.oldValue)
    newStr = str(self.newValue)
    info = infoSpec % (ownerName, fieldName, oldStr, newStr)
    from worktoy.utilities import textFmt
    return textFmt(info)

  __repr__ = __str__



"""
ELEMENTSIZE provides a sentinel used by the AbstractTensor class to indicate
the location of the element size in bytes.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class ELEMENTSIZE(Sentinel):
  """
  ELEMENTSIZE provides a sentinel used by the AbstractTensor class to
  indicate the location of the element size in bytes.
  """



"""
WithoutException provides a custom exception raised to indicate that a
context-only method called without descriptor-context.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.utilities import maybe

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class WithoutException(RuntimeError):
  """
  WithoutException provides a custom exception raised to indicate a call
  without context made to a context-only method.
  """

  __slots__ = ('desc',)

  def __init__(self, desc: Any) -> None:
    self.desc = desc
    RuntimeError.__init__(self, )

  def __str__(self, ) -> str:
    infoSpec = """Context-less call detected for descriptor at '%s.%s'!"""
    fbOwner = 'NO-OWNER'
    ownerName = getattr(self.desc.getFieldOwner(), '__name__', fbOwner)
    fbName = 'NO-FIELD'
    fieldName = maybe(self.desc.getFieldName(), fbName)
    info = infoSpec % (ownerName, fieldName)
    return info

  __repr__ = __str__



"""
TypeCast encapsulates the logic for instantiating a type from arguments.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.waitaminute.dispatch import TypeCastException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


def typeCast(target: type, arg: Any) -> Any:
  """
  Casts the given argument to the specified type, if supported.
  """
  if isinstance(arg, target):
    return target(arg)
  if target is str:
    if isinstance(arg, (bytes, bytearray)):
      return arg.decode('utf-8')
    try:
      flattened = arg[0]
    except Exception as exception:
      raise TypeCastException(target, arg) from exception
    else:
      return str(flattened)
  if target is int:
    if isinstance(arg, complex):
      if arg.imag:
        raise TypeCastException(target, arg)
      return typeCast(target, arg.real)
    if isinstance(arg, float):
      if float.is_integer(arg):
        return int(arg)
      raise TypeCastException(target, arg)
    if isinstance(arg, str):
      try:
        castArg = int(arg)
      except Exception as exception:
        raise TypeCastException(target, arg) from exception
      else:
        return castArg
  if target is float:
    if isinstance(arg, int):
      return float(arg)
    if isinstance(arg, complex):
      if arg.imag:
        raise TypeCastException(target, arg)
      return float(arg.real)
    if isinstance(arg, str):
      try:
        castArg = float(arg)
      except Exception as exception:
        raise TypeCastException(target, arg) from exception
      else:
        return castArg
  if target is complex:
    if isinstance(arg, (int, float)):
      return float(arg) + 0j
  if target is bool:
    return True if arg else False
  if target in [list, tuple, set, frozenset, dict]:
    raise TypeCastException(target, arg)
  try:
    castArg = target(arg)
  except Exception as exception:
    if isinstance(arg, str):
      raise TypeCastException(target, arg) from exception
    try:
      arg = arg[0]
    except Exception:
      raise TypeCastException(target, arg) from exception
    else:
      try:
        out = target(arg)
      except Exception:
        raise TypeCastException(target, arg) from exception
      else:
        return out
  else:
    return castArg



"""
AccessError is raised when a descriptor does not provide a way to retrieve
a requested value. This exception is raised by 'Desc' and its subclasses
failing to implement '__instance_get__' or which falls back to the parent
implementation under certain conditions.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import DescriptorException
from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class AccessError(DescriptorException, AttributeError):
  """
  AccessError is raised when a descriptor does not provide a way to retrieve
  a requested value. This exception is raised by 'Desc' and its subclasses
  failing to implement '__instance_get__' or which falls back to the parent
  implementation under certain conditions.
  """

  __slots__ = ('desc',)

  def __init__(self, desc) -> None:
    """
    Initializes the AccessError with the class that failed to provide a
    value.
    """
    self.desc = desc
    DescriptorException.__init__(self, )

  def __str__(self) -> str:
    """
    Returns a string representation of the AccessError.
    """
    infoSpec = """The '%s' descriptor at '%s.%s' failed to retrieve a 
    value!"""
    descTypeName = type(self.desc).__name__
    ownerName = self.desc.__field_owner__.__name__
    fieldName = self.desc.__field_name__
    info = infoSpec % (descTypeName, ownerName, fieldName)
    return textFmt(info)

  __repr__ = __str__



"""
Sentinels represent special situations. The sentinel itself provides no
information about its meaning or function. It is merely a stateless unique
object equal only to itself.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._sentinel import Sentinel
from ._deleted import DELETED
from ._owner import OWNER
from ._this import THIS
from ._desc import DESC
from ._locked import LOCKED
from ._element_size import ELEMENTSIZE

__all__ = [
    'Sentinel',
    'DELETED',
    'OWNER',
    'THIS',
    'DESC',
    'LOCKED',
    'ELEMENTSIZE',
]



"""MissingVariable exception should be raised when a variable is missing
and requires initialization. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


class MissingVariable(Exception):
  """MissingVariable exception should be raised when a variable is missing
  and requires initialization. """

  __slots__ = ('varName', 'varType')

  def __init__(self, name: str, type_: type) -> None:
    """Initialize the MissingVariable object."""
    self.varName = name
    self.varType = type_
    Exception.__init__(self, )

  def __str__(self) -> str:
    """Return the string representation of the MissingVariable."""
    infoSpec = """Missing variable at name: '%s' of type: '%s'!"""
    name = self.varName
    clscName = self.varType.__name__
    info = infoSpec % (name, clscName)
    from ..utilities import textFmt
    return textFmt(info)

  __repr__ = __str__



"""
Dir provides a descriptor identifying the file from which the owner
received by __get__ is imported.

Please note, that the file is returned
only when the descriptor is accessed through an instance. If accessed
through the class, the descriptor object returns itself.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import sys
import os
from typing import TYPE_CHECKING

from worktoy.waitaminute.desc import ReadOnlyError, ProtectedError

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Never


class Directory:
  """
  A descriptor that returns the directory of the file containing the
  instance accessing it.
  """

  def __get__(self, instance: Any, owner: type) -> Any:
    """
    Returns the directory of the file containing the instance accessing it.
    """
    if instance is None:
      return self
    module = sys.modules.get(instance.__class__.__module__)
    filePath = getattr(module, '__file__')
    return os.path.abspath(os.path.dirname(filePath))

  def __set__(self, instance: Any, value: Any, **kwargs) -> Never:
    raise ReadOnlyError(instance, self, value, )

  def __delete__(self, instance: Any, **kwargs) -> Never:
    raise ProtectedError(self, instance, )



"""
KeeMeta provides the metaclass creating the KeeNum enumeration class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls import AbstractMetaclass
from ..utilities import textFmt
from ..waitaminute import TypeException
from ..waitaminute.meta import IllegalInstantiation
from ..waitaminute.keenum import EmptyKeeNumError
from . import KeeSpace, _KeeNumBase

if TYPE_CHECKING:  # pragma: no cover
  from typing import Self, TypeAlias, Any, Iterator

  Bases: TypeAlias = tuple[type, ...]
  MTypes: TypeAlias = dict[str, type]


class KeeMeta(AbstractMetaclass):
  """
  KeeMeta provides the metaclass creating the KeeNum enumeration class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __core_keenum__ = None  # Future KeeNum base class

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _createCoreKeeNum(mcls, ) -> Self:
    """
    Creator function for the core KeeNum class.
    """

    class KeeNum(_KeeNumBase, metaclass=mcls):
      """
      KeeNum provides the base class for enumerating classes with
      restricted and predefined instances called members.
      """
      pass

    setattr(mcls, '__core_keenum__', KeeNum)

  @classmethod
  def getCoreKeeNum(mcls, **kwargs) -> type:
    """
    Get the core KeeNum class.
    """
    if mcls.__core_keenum__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError  # pragma: no cover
      mcls._createCoreKeeNum()
      return mcls.getCoreKeeNum(_recursion=True)
    return mcls.__core_keenum__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> KeeSpace:
    """
    Prepare the class namespace.
    """
    return KeeSpace(mcls, name, bases, **kwargs)

  def __new__(mcls, name: str, bases: Bases, space: KeeSpace, **kw) -> Self:
    """
    Create a new instance of the KeeMeta metaclass.
    """
    if name == 'KeeNum' or kw.get('_root', False):
      return super().__new__(mcls, name, bases, space, **kw)
    coreKeeNum = mcls.getCoreKeeNum()
    return super().__new__(mcls, name, (coreKeeNum,), space, **kw)

  def __init__(cls, name: str, bases: Bases, space: KeeSpace, **kw) -> None:
    """
    Initialize the KeeMeta metaclass.
    """
    super().__init__(name, bases, space, **kw)
    if name == 'KeeNum':
      return
    futureEntries = getattr(cls, '__future_entries__', )
    actualEntries = dict()
    valueType = getattr(cls, '__future_value_type__', None)
    if valueType is None:
      raise EmptyKeeNumError(name)
    memberValues = []
    memberNames = []
    for i, (name, value) in enumerate(futureEntries.items()):
      entry = cls(_root=True)
      entry.__member_name__ = name
      entry.__member_value__ = value
      entry.__member_index__ = i
      memberValues.append(value)
      memberNames.append(name)
      actualEntries[name] = entry
      setattr(cls, name, entry)

    setattr(cls, '__value_type__', valueType)
    setattr(cls, '__member_entries__', actualEntries)
    delattr(cls, '__future_entries__')
    delattr(cls, '__future_value_type__')

  def __call__(cls, *args, **kwargs) -> Any:
    """
    Only the _addMember method is allowed to create instances of the
    KeeNum class.
    """
    if not kwargs.get('_root', False):
      try:
        member = cls.__getitem__(args[0])
      except Exception as exception:
        raise IllegalInstantiation(cls) from exception
      else:
        return member
    self = super().__call__(**kwargs)
    return self

  def __len__(cls) -> int:
    """
    Get the number of members in the KeeNum class.
    """
    out = 0
    for _ in cls:
      out += 1
    else:
      return out

  def __iter__(cls, ) -> Iterator:
    """
    Iterate over the members of the KeeNum class.
    """
    yield from getattr(cls, '__member_entries__', ).values()

  def __getitem__(cls, key: str) -> Any:
    """
    Get a member by its name.
    """
    valueType = getattr(cls, '__value_type__', None)
    if isinstance(key, int):
      return cls._getFromIndex(key)
    if isinstance(key, str):
      return cls._getFromName(key)
    if isinstance(key, valueType):
      return cls._getFromValue(key)
    raise TypeException('key', key, int, str, valueType)

  def _getFromIndex(cls, index: int) -> Any:
    """
    Get a member by its index. If the value type is 'int' and the index is
    negative, this method assumes that the index does not match any value
    type.
    """
    if index < 0:
      return cls._getFromIndex(len(cls) + index)
    for entry in cls:
      if entry.index == index:
        return entry
    infoSpec = """Index %d out of range for %s"""
    info = infoSpec % (index, cls.__name__)
    raise IndexError(textFmt(info))

  def _getFromName(cls, name: str, **kwargs) -> Any:
    """
    Get a member by its name. If the value type is 'str', this method
    assumes that the identifier does not match any value type.
    """
    for entry in cls:
      if entry.name == name:
        return entry
    for entry in cls:
      entryName = entry.name.lower().replace('_', '')
      lookupName = name.lower().replace('_', '')
      if entryName == lookupName:
        return entry
    raise KeyError(name)

  def _getFromValue(cls, value: Any, **kwargs) -> Any:
    """
    Get a member by its value.
    """
    for entry in cls:
      if entry.value == value:
        return entry
    raise ValueError(value)

  def __str__(cls) -> str:
    """
    String representation of the KeeMeta metaclass.
    """
    if cls.__name__ == 'KeeNum':
      return """<class 'KeeNum'>"""
    valueType = getattr(cls, '__value_type__', )
    typeName = '' if valueType is str else valueType.__name__
    clsName = cls.__name__
    infoSpec = """%s(KeeNum)[%s]"""
    return infoSpec % (clsName, typeName)

  __repr__ = __str__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



"""
The 'worktoy.waitaminute.dispatch' module provides custom exceptions
specifically related to the overload control flow.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._dispatch_exception import DispatchException
from ._hash_mismatch import HashMismatch
from ._cast_mismatch import CastMismatch
from ._flex_mismatch import FlexMismatch
from ._type_cast_exception import TypeCastException

___all__ = [
    'DispatchException',
    'HashMismatch',
    'CastMismatch',
    'FlexMismatch',
    'TypeCastException',
]



"""
Provides the exponential function and natural logarithm
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import cos, sin, atan2

if TYPE_CHECKING:  # pragma: no cover
  from typing import Union


def _exp(z: float, f: float = 1.0) -> float:
  """Returns the exponential of z using the Taylor series expansion."""
  if z > 2:
    return _exp(z / 2, f * 2)
  if z < 0:
    return 1 / _exp(-z)
  shift = 1e16
  term = 1 / shift
  result = 1 / shift
  n = 1

  while abs(term) > 1e-48:
    term *= z / n
    result += term
    n += 1
  else:
    return (shift * result) ** f


def _e() -> float:
  """Returns the base of the natural logarithm."""
  return _exp(1)


e = _e()


def _log(x: Union[float, complex]) -> Union[float, complex]:
  if x ** 2 < 1e-16:
    raise ZeroDivisionError
  if x < 1:
    return -_log(1 / x)
  if x > e:
    return _log(x / e) + 1
  if (1 - x) ** 2 < 1e-32:
    return 0
  term = (x - 1) / x
  result = term
  n = 1

  while abs(term) > 1e-32:
    term *= (x - 1) / x
    result += term / (n + 1)
    n += 1
  else:
    return result


def exp(z: complex) -> complex:
  """Returns the exponential of a complex number."""
  if isinstance(z, (int, float)):
    return exp(z * (1.0 + 0j))
  r = _exp(z.real)
  return r * (cos(z.imag) + sin(z.imag) * 1j)


def log(z: complex, ) -> complex:
  r = abs(z)
  t = atan2(z.imag, z.real)
  return _log(r) + t * 1j



"""The 'scrapDirectory' function removes empty directories."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from ..utilities import textFmt
from . import validateExistingDirectory

if TYPE_CHECKING:  # pragma: no cover
  pass


def scrapDirectory(dirPath: str, **kwargs) -> None:
  """
  Removes empty directory at the specified path.
  """
  try:
    validateExistingDirectory(dirPath)
  except FileNotFoundError as fileNotFoundError:
    if kwargs.get('strict', True):
      raise fileNotFoundError
  except NotADirectoryError as notADirectoryError:
    infoSpec = """The path received by 'scrapDirectory': '%s' is not a 
    directory!"""
    info = textFmt(infoSpec % dirPath)
    raise NotADirectoryError(info) from notADirectoryError
  else:
    os.rmdir(dirPath)



"""ReadOnlyError is raised when an attempt is made to modify a read-only
attribute. This is a subclass of TypeError and should be used to indicate
that the attribute is read-only. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import DescriptorException
from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class ReadOnlyError(DescriptorException):
  """ReadOnlyError is raised when an attempt is made to modify a read-only
  attribute. This is a subclass of TypeError and should be used to indicate
  that the attribute is read-only. """

  __slots__ = ('instance', 'desc', 'newVal',)

  def __init__(self, instance: Any, desc: Any, val: Any, ) -> None:
    """Initialize the ReadOnlyError."""
    self.instance = instance
    self.desc = desc
    self.newVal = val
    DescriptorException.__init__(self, )

  def __str__(self, ) -> str:
    """Return the string representation of the ReadOnlyError."""
    infoSpec = """Attempted to overwrite read-only attribute '%s' with 
    new value: '%s'!"""
    ownerName = type(self.instance).__name__
    fieldName = getattr(self.desc, '__field_name__', 'object')
    fieldId = '%s.%s' % (ownerName, fieldName)
    newValue = str(self.newVal)
    info = infoSpec % (fieldId, newValue)
    return textFmt(info)

  __repr__ = __str__



"""PathSyntaxException provides a custom exception raised to indicate that
a 'str' object is not a valid absolute path. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Union, TypeAlias, LiteralString

  Path: TypeAlias = Union[str, bytes, LiteralString]


class PathSyntaxException(ValueError):
  """
  PathSyntaxException provides a custom exception raised to indicate that
  a 'str' object is not a valid absolute path.
  """

  __slots__ = ('badPath',)

  def __init__(self, path: Path) -> None:
    """
    Initialize the PathSyntaxException with the invalid path.

    Args:
      path (str): The invalid path.
    """
    self.badPath = path
    ValueError.__init__(self, )

  def __str__(self, ) -> str:
    """
    Get the string representation of the exception.
    """

    infoSpec = """The path '%s' is not a valid, absolute path!"""
    return textFmt(infoSpec % self.badPath)

  __repr__ = __str__



"""
Alias provides a descriptor allowing renaming of a descriptor, typically
one inherited from a parent.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core import Object

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Type


class Alias(Object):
  """
  Alias provides a descriptor allowing renaming of a descriptor, typically
  one inherited from a parent.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __real_name__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type) -> Any:
    realDesc = getattr(type(instance), self.__real_name__, )
    return realDesc.__get__(instance, owner)

  def __set__(self, instance: Any, value: Any, **kwargs) -> None:
    """
    Sets the value of the aliased descriptor.
    """
    realDesc = getattr(type(instance), self.__real_name__, )
    return realDesc.__set__(instance, value, **kwargs)

  def __delete__(self, instance: Any, **kwargs) -> None:
    """
    Deletes the value of the aliased descriptor.
    """
    realDesc = getattr(type(instance), self.__real_name__, )
    return realDesc.__delete__(instance, **kwargs)

  def __set_name__(self, owner: Type[Object], name: str) -> None:
    try:
      _ = getattr(owner, self.__real_name__)
    except AttributeError as attributeError:
      raise RuntimeError from attributeError
    else:
      Object.__set_name__(self, owner, name)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, realName: str) -> None:
    """
    Initializes the Alias descriptor with the name of the real descriptor.
    """
    self.__real_name__ = realName



"""
DELETED indicates deletion. If any 'object.__getattribute__' would return
DELETED, it should instead raise an AttributeError. The core classes in
the 'worktoy.core' module implements this behaviour. The example below
illustrates the motivation behind this sentinel.

A descriptor class implements a default value that descriptor objects
return if the normal __get__ is not able to find a value on a given
owning instance. Each descriptor object may have its own default value,
but if not, the class provides a fallback value. Thus, when accessing
the descriptor described in this example, the descriptor attempts three
separate lookups:
1.  from the instance passed to __get__
2.  a default value set on the descriptor object itself
3.  the fallback value set on the descriptor class itself

Instead of requiring handling three different lookups, the '__delete__'
functionality may be implemented by 'setting' the descriptor to the
DELETED sentinel. Then, the first lookup finds the DELETED sentinel,
which signals the 'Desc.__get__' method to raise 'AttributeError'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class DELETED(Sentinel):
  """
  DELETED should raise an AttributeError
  """
  pass



"""
DelException is a custom exception raised when someone attempts to create
a class that implements the '__del__' method without providing the custom
keyword argument: 'trustMeBro=True'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


class DelException(SyntaxError):
  """
  DelException is a custom exception raised when someone attempts to create
  a class that implements the '__del__' method without providing the custom
  keyword argument: 'trustMeBro=True'.
  """
  __slots__ = ('mcls', 'name', 'bases', 'space')

  def __init__(self, *args) -> None:
    """Initialize the DelException with the class."""
    self.mcls, self.name, self.bases, self.space = args
    SyntaxError.__init__(self, )

  def __str__(self) -> str:
    """Return a string representation of the DelException."""
    infoSpec = """When attempting to derive a class named '%s' from the 
    metaclass '%s', the '__del__' method was found in the namespace! This 
    is almost always a typo, but if not this error can be suppressed by 
    passing the keyword argument 'trustMeBro=True' during class creation. """
    if TYPE_CHECKING:  # pragma: no cover
      assert isinstance(self.bases, tuple)
    if self.bases:
      mclsSpec = """%s with bases: (%s)"""
    else:
      mclsSpec = """%s%s"""
    basesStr = ', '.join(base.__name__ for base in self.bases)
    mclsName = mclsSpec % (self.mcls.__name__, basesStr)
    info = infoSpec % (self.name, mclsName)
    return info

  __repr__ = __str__



"""
The 'attributeErrorFactory' instantiates 'AttributeError' with a message
that matches the built-in message if a given object does not have a given
attribute. This allows simulating the absense of an attribute. For example:

class Bar:  # secret descriptor!

  __fallback_value__ = 'lol'
  __field_name__ = None
  __field_owner__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: object, owner: type) -> object:
    if instance is None:
      return self
    if hasattr(owner, '__trust_me_bro__'):
      pvtName = '__%s__' % self.__field_name__
      fallback = self.__fallback_value__
      return getattr(instance, pvtName, fallback)
    raise attributeErrorFactory(owner, self.__field_name__)


class Foo:
  __trust_me_bro__ = True
  bar = Bar()


class Sus:
  bar = Bar()


if __name__ == '__main__':
  foo = Foo()
  print(foo.bar)  # This will return the fallback value 'lol'
  sus = Sus()
  print(sus.bar)  # Raises AttributeError


The above attribute error will have the following message:
AttributeError: 'Sus' object has no attribute 'bar'
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


def attributeErrorFactory(owner: str, field: str) -> AttributeError:
  """
  Factory function that creates an AttributeError with a message that
  matches the built-in message if a given object does not have a given
  attribute.
  """
  infoSpec = """AttributeError: '%s' object has no attribute '%s'"""
  info = infoSpec % (str(owner), str(field))
  return AttributeError(info)  # type: ignore[return-value]
